# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.circle import Circle
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Vec3, Matrix44, OCS, NonUniformScalingError

TEST_CLASS = Circle
TEST_TYPE = "CIRCLE"

ENTITY_R12 = """0
CIRCLE
5
0
8
0
10
0.0
20
0.0
30
0.0
40
1.0
"""

ENTITY_R2000 = """0
CIRCLE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbCircle
10
0.0
20
0.0
30
0.0
40
1.0
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_negative_radius():
    circle = Circle.new(dxfattribs={"radius": -1})
    assert circle.dxf.radius == -1, "Radius < 0 is valid"


def test_zero_radius():
    circle = Circle.new(dxfattribs={"radius": 0})
    assert circle.dxf.radius == 0, "Radius == 0 is valid"


def test_extrusion_can_not_be_a_null_vector():
    circle = Circle.new(dxfattribs={"extrusion": (0, 0, 0)})
    assert circle.dxf.extrusion == (0, 0, 1), "expected default extrusion"


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "center": (1, 2, 3),
            "radius": 2.5,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"

    assert entity.dxf.center == (1, 2, 3)
    assert entity.dxf.center.x == 1, "is not Vec3 compatible"
    assert entity.dxf.center.y == 2, "is not Vec3 compatible"
    assert entity.dxf.center.z == 3, "is not Vec3 compatible"
    assert entity.dxf.radius == 2.5
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1
    assert entity.dxf.extrusion == (0.0, 0.0, 1.0)
    assert entity.dxf.hasattr("extrusion") is False, "just the default value"


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.center == (0, 0, 0)
    assert entity.dxf.radius == 1


def test_get_point_2d_circle():
    radius = 2.5
    z = 3.0
    circle = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "center": (1, 2, z),
            "radius": radius,
        },
    )
    vertices = list(circle.vertices([90]))
    assert vertices[0].isclose(Vec3(1, 2 + radius, z))


def test_get_point_with_ocs():
    circle = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "center": (1, 2, 3),
            "radius": 2.5,
            "extrusion": (0, 0, -1),
        },
    )
    vertices = list(circle.vertices([90, 180]))
    assert vertices[0].isclose(Vec3(-1, 4.5, -3), abs_tol=1e-6)
    assert vertices[1].isclose(Vec3(1.5, 2, -3), abs_tol=1e-6)


@pytest.mark.parametrize(
    "txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)]
)
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    circle = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    circle.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    circle.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_circle_default_ocs():
    circle = Circle.new(dxfattribs={"center": (2, 3, 4), "thickness": 2})
    # 1. rotation - 2. scaling - 3. translation
    m = Matrix44.chain(Matrix44.scale(2, 2, 3), Matrix44.translate(1, 1, 1))
    # default extrusion is (0, 0, 1), therefore scale(2, 2, ..) is a uniform
    # scaling in the xy-play of the OCS
    circle.transform(m)

    assert circle.dxf.center == (5, 7, 13)
    assert circle.dxf.extrusion == (0, 0, 1)
    assert circle.dxf.thickness == 6


def test_circle_fast_translation():
    circle = Circle.new(
        dxfattribs={"center": (2, 3, 4), "extrusion": Vec3.random()}
    )
    ocs = circle.ocs()
    offset = Vec3(1, 2, 3)
    center = ocs.to_wcs(circle.dxf.center) + offset
    circle.translate(*offset)
    assert ocs.to_wcs(circle.dxf.center).isclose(center, abs_tol=1e-9)


def test_circle_non_uniform_scaling():
    circle = Circle.new(
        dxfattribs={"center": (2, 3, 4), "extrusion": (0, 1, 0), "thickness": 2}
    )
    # extrusion in WCS y-axis, therefore scale(2, ..., 3) is a non uniform
    # scaling in the xy-play of the OCS which is the xz-plane of the WCS
    with pytest.raises(NonUniformScalingError):
        circle.transform(Matrix44.scale(2, 2, 3))

    # source values unchanged after exception
    assert circle.dxf.center == (2, 3, 4)
    assert circle.dxf.extrusion == (0, 1, 0)
    assert circle.dxf.thickness == 2


def test_circle_user_ocs():
    center = (2, 3, 4)
    extrusion = (0, 1, 0)

    circle = Circle.new(
        dxfattribs={"center": center, "extrusion": extrusion, "thickness": 2}
    )
    ocs = OCS(extrusion)
    v = ocs.to_wcs(center)  # (-2, 4, 3)
    v = Vec3(v.x * 2, v.y * 4, v.z * 2)
    v += (1, 1, 1)
    # and back to OCS, extrusion is unchanged
    result = ocs.from_wcs(v)

    m = Matrix44.chain(Matrix44.scale(2, 4, 2), Matrix44.translate(1, 1, 1))
    circle.transform(m)
    assert circle.dxf.center == result
    assert circle.dxf.extrusion == (0, 1, 0)
    assert circle.dxf.thickness == 8  # in WCS y-axis


@pytest.mark.parametrize(
    "radius, sagitta, count",
    [
        (1, 0.35, 5),
        (1, 0.10, 8),
        (0, 0.35, 0),
        (0, 0.10, 0),  # radius 0 works but yields nothing
        (-1, 0.35, 5),
        (-1, 0.10, 8),  # negative radius same as positive radius
    ],
)
def test_circle_flattening(radius, sagitta, count):
    circle = Circle.new(dxfattribs={"radius": radius})
    assert len(list(circle.flattening(sagitta))) == count


MALFORMED_CIRCLE = """0
CIRCLE
5
0
62
7
330
0
6
LT_EZDXF
8
LY_EZDXF
100
AcDbCircle
10
1.0
20
2.0
30
3.0
100
AcDbEntity
40
2.0
"""


def test_load_malformed_circle():
    circle = Circle.from_text(MALFORMED_CIRCLE)
    assert circle.dxf.layer == "LY_EZDXF"
    assert circle.dxf.linetype == "LT_EZDXF"
    assert circle.dxf.color == 7
    assert circle.dxf.center.isclose((1, 2, 3))
    assert circle.dxf.radius == 2.0



# Copyright (c) 2018-2023 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterator,
    cast,
    Optional,
    TypeVar,
    Generic,
)
from ezdxf.lldxf.const import (
    DXFValueError,
    DXFKeyError,
    INVALID_NAME_CHARACTERS,
)
from ezdxf.lldxf.validator import make_table_key, is_valid_table_name

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFObject, Dictionary


def validate_name(name: str) -> str:
    name = name[:255]
    if not is_valid_table_name(name):
        raise DXFValueError(
            f"table name '{name}' contains invalid characters: {INVALID_NAME_CHARACTERS}"
        )
    return name


T = TypeVar("T", bound="DXFObject")


class ObjectCollection(Generic[T]):
    """
    Note:
    ObjectCollections may contain entries where the name stored in the entity as
    "name" attribute diverges from the key in the DICTIONARY object e.g. MLEADERSTYLE
    collection may have entries for "Standard" and "Annotative" but both MLEADERSTYLE
    objects have the name "Standard".

    """

    def __init__(
        self,
        doc: Drawing,
        dict_name: str = "ACAD_MATERIAL",
        object_type: str = "MATERIAL",
    ):
        self.doc: Drawing = doc
        self.object_dict_name = dict_name
        self.object_type: str = object_type
        self.object_dict: Dictionary = doc.rootdict.get_required_dict(dict_name)

    def update_object_dict(self) -> None:
        self.object_dict = self.doc.rootdict.get_required_dict(self.object_dict_name)

    def create_required_entries(self) -> None:
        pass

    def __iter__(self) -> Iterator[tuple[str, T]]:
        return self.object_dict.items()

    def __len__(self) -> int:
        return len(self.object_dict)

    def __contains__(self, name: str) -> bool:
        return self.has_entry(name)

    def __getitem__(self, name: str) -> T:
        entry = self.get(name)
        if entry is None:
            raise DXFKeyError(name)
        return entry

    @property
    def handle(self) -> str:
        """Returns the DXF handle of the DICTIONARY object."""
        return self.object_dict.dxf.handle

    @property
    def is_hard_owner(self) -> bool:
        """Returns ``True`` if the collection is hard owner of entities.
        Hard owned entities will be destroyed by deleting the dictionary.
        """
        return self.object_dict.is_hard_owner

    def has_entry(self, name: str) -> bool:
        return self.get(name) is not None

    def is_unique_name(self, name: str) -> bool:
        name = make_table_key(name)
        for entry_name in self.object_dict.keys():
            if make_table_key(entry_name) == name:
                return False
        return True

    def get(self, name: str, default: Optional[T] = None) -> Optional[T]:
        """Get object by name. Object collection entries are case-insensitive.

        Args:
            name: object name as string
            default: default value

        """
        name = make_table_key(name)
        for entry_name, obj in self.object_dict.items():
            if make_table_key(entry_name) == name:
                return obj
        return default

    def new(self, name: str) -> T:
        """Create a new object of type `self.object_type` and store its handle
        in the object manager dictionary.  Object collection entry names are
        case-insensitive and limited to 255 characters.

        Args:
            name: name of new object as string

        Returns:
            new object of type `self.object_type`

        Raises:
            DXFValueError: if object name already exist or is invalid

        (internal API)

        """
        name = validate_name(name)
        if not self.is_unique_name(name):
            raise DXFValueError(f"{self.object_type} entry {name} already exists.")
        return self._new(name, dxfattribs={"name": name})

    def duplicate_entry(self, name: str, new_name: str) -> T:
        """Returns a new table entry `new_name` as copy of `name`,
        replaces entry `new_name` if already exist.

        Raises:
             DXFValueError: `name` does not exist

        """
        entry = self.get(name)
        if entry is None:
            raise DXFValueError(f"entry '{name}' does not exist")
        new_name = validate_name(new_name)
        # remove existing entry
        existing_entry = self.get(new_name)
        if existing_entry is not None:
            self.delete(new_name)

        entitydb = self.doc.entitydb
        if entitydb:
            new_entry = entitydb.duplicate_entity(entry)
        else:  # only for testing!
            new_entry = entry.copy()
        if new_entry.dxf.is_supported("name"):
            new_entry.dxf.name = new_name
        self.object_dict.add(new_name, new_entry)  # type: ignore
        owner_handle = self.object_dict.dxf.handle
        new_entry.dxf.owner = owner_handle
        new_entry.set_reactors([owner_handle])
        return new_entry  # type: ignore

    def _new(self, name: str, dxfattribs: dict) -> T:
        objects = self.doc.objects
        assert objects is not None

        owner = self.object_dict.dxf.handle
        dxfattribs["owner"] = owner
        obj = objects.add_dxf_object_with_reactor(
            self.object_type, dxfattribs=dxfattribs
        )
        self.object_dict.add(name, obj)
        return cast(T, obj)

    def delete(self, name: str) -> None:
        objects = self.doc.objects
        assert objects is not None

        obj = self.get(name)  # case insensitive
        if obj is not None:
            # The underlying DICTIONARY is not case-insensitive implemented,
            # get real object name if available
            if obj.dxf.is_supported("name"):
                name = obj.dxf.get("name", name)
            self.object_dict.discard(name)
            objects.delete_entity(obj)

    def clear(self) -> None:
        """Delete all entries."""
        self.object_dict.clear()



# Copyright (c) 2021-2024, Matthew Broadway
# License: MIT License
from __future__ import annotations
from typing import Optional
import warnings
import dataclasses
from dataclasses import dataclass
from enum import Enum, auto

from ezdxf import disassemble
from ezdxf.enums import Measurement
from .type_hints import Color


class LinePolicy(Enum):
    """This enum is used to define how to render linetypes.

    .. note::

        Text and shapes in linetypes are not supported.

    Attributes:
        SOLID: draw all lines as solid regardless of the linetype style
        ACCURATE: render styled lines as accurately as possible
        APPROXIMATE: ignored since v0.18.1 - uses always ACCURATE by default

    """

    SOLID = auto()
    APPROXIMATE = auto()  # ignored since v0.18.1
    ACCURATE = auto()


class ProxyGraphicPolicy(Enum):
    """The action to take when an entity with a proxy graphic is encountered

    .. note::

        To get proxy graphics support proxy graphics have to be loaded:
        Set the global option :attr:`ezdxf.options.load_proxy_graphics` to
        ``True``, which is the default value.

        This can not prevent drawing proxy graphic inside of blocks,
        because this is beyond the domain of the drawing add-on!

    Attributes:
        IGNORE: do not display proxy graphics (skip_entity will be called instead)
        SHOW: if the entity cannot be rendered directly (e.g. if not implemented)
            but a proxy is present: display the proxy
        PREFER: display proxy graphics even for entities where direct rendering
            is available
    """

    IGNORE = auto()
    SHOW = auto()
    PREFER = auto()


class HatchPolicy(Enum):
    """The action to take when a HATCH entity is encountered

    Attributes:
        NORMAL: render pattern and solid fillings
        IGNORE: do not show HATCH entities at all
        SHOW_OUTLINE: show only the outline of HATCH entities
        SHOW_SOLID: show HATCH entities as solid filling regardless of the pattern

    """

    NORMAL = auto()
    IGNORE = auto()
    SHOW_OUTLINE = auto()
    SHOW_SOLID = auto()
    SHOW_APPROXIMATE_PATTERN = auto()  # ignored since v0.18.1 == NORMAL


class LineweightPolicy(Enum):
    """This enum is used to define how to determine the lineweight.

    Attributes:
        ABSOLUTE: in mm as resolved by the :class:`Frontend` class
        RELATIVE: lineweight is relative to page size
        RELATIVE_FIXED: fixed lineweight relative to page size for all strokes

    """

    ABSOLUTE = auto()
    # set fixed lineweight for all strokes in absolute mode:
    # set Configuration.min_lineweight to the desired lineweight in 1/300 inch!
    # set Configuration.lineweight_scaling to 0

    # The RELATIVE policy is a backend feature and is not supported by all backends!
    RELATIVE = auto()
    RELATIVE_FIXED = auto()


class ColorPolicy(Enum):
    """This enum is used to define how to determine the line/fill color.

    Attributes:
        COLOR: as resolved by the :class:`Frontend` class
        COLOR_SWAP_BW: as resolved by the :class:`Frontend` class but swaps black and white
        COLOR_NEGATIVE: invert all colors
        MONOCHROME: maps all colors to gray scale in range [0%, 100%]
        MONOCHROME_DARK_BG: maps all colors to gray scale in range [30%, 100%], brightens
            colors for dark backgrounds
        MONOCHROME_LIGHT_BG:  maps all colors to gray scale in range [0%, 70%], darkens
            colors for light backgrounds
        BLACK: maps all colors to black
        WHITE: maps all colors to white
        CUSTOM: maps all colors to custom color :attr:`Configuration.custom_fg_color`

    """

    COLOR = auto()
    COLOR_SWAP_BW = auto()
    COLOR_NEGATIVE = auto()
    MONOCHROME = auto()
    MONOCHROME_DARK_BG = auto()
    MONOCHROME_LIGHT_BG = auto()
    BLACK = auto()
    WHITE = auto()
    CUSTOM = auto()


class BackgroundPolicy(Enum):
    """This enum is used to define the background color.

    Attributes:
        DEFAULT: as resolved by the :class:`Frontend` class
        WHITE: white background
        BLACK: black background
        PAPERSPACE: default paperspace background
        MODELSPACE: default modelspace background
        OFF: fully transparent background
        CUSTOM: custom background color by :attr:`Configuration.custom_bg_color`

    """

    DEFAULT = auto()
    WHITE = auto()
    BLACK = auto()
    PAPERSPACE = auto()
    MODELSPACE = auto()
    OFF = auto()
    CUSTOM = auto()


class TextPolicy(Enum):
    """This enum is used to define the text rendering.

    Attributes:
        FILLING: text is rendered as solid filling (default)
        OUTLINE: text is rendered as outline paths
        REPLACE_RECT: replace text by a rectangle
        REPLACE_FILL: replace text by a filled rectangle
        IGNORE: ignore text entirely

    """

    FILLING = auto()
    OUTLINE = auto()
    REPLACE_RECT = auto()
    REPLACE_FILL = auto()
    IGNORE = auto()


class ImagePolicy(Enum):
    """This enum is used to define the image rendering.

    Attributes:
        DISPLAY: display images as they would appear in a regular CAD application
        RECT: display images as rectangles
        MISSING: images are always rendered as-if they are missing (rectangle + path text)
        PROXY: images are rendered using their proxy representations (rectangle)
        IGNORE: ignore images entirely

    """

    DISPLAY = auto()
    RECT = auto()
    MISSING = auto()
    PROXY = auto()
    IGNORE = auto()


@dataclass(frozen=True)
class Configuration:
    """Configuration options for the :mod:`drawing` add-on.

    Attributes:
        pdsize: the size to draw POINT entities (in drawing units)
            set to None to use the $PDSIZE value from the dxf document header

            ======= ====================================================
            0       5% of draw area height
            <0      Specifies a percentage of the viewport size
            >0      Specifies an absolute size
            None    use the $PDMODE value from the dxf document header
            ======= ====================================================

        pdmode: point styling mode (see POINT documentation)

            see :class:`~ezdxf.entities.Point` class documentation

        measurement: whether to use metric or imperial units as enum :class:`ezdxf.enums.Measurement`

            ======= ======================================================
            0       use imperial units (in, ft, yd, ...)
            1       use metric units (ISO meters)
            None    use the $MEASUREMENT value from the dxf document header
            ======= ======================================================

        show_defpoints: whether to show or filter out POINT entities on the defpoints layer
        proxy_graphic_policy: the action to take when a proxy graphic is encountered
        line_policy: the method to use when drawing styled lines (eg dashed,
            dotted etc)
        hatch_policy: the method to use when drawing HATCH entities
        infinite_line_length: the length to use when drawing infinite lines
        lineweight_scaling:
            multiplies every lineweight by this factor; set this factor to 0.0 for a
            constant minimum line width defined by the :attr:`min_lineweight` setting
            for all lineweights;
            the correct DXF lineweight often looks too thick in SVG, so setting a
            factor < 1 can improve the visual appearance
        min_lineweight: the minimum line width in 1/300 inch; set to ``None`` for
            let the backend choose.
        min_dash_length: the minimum length for a dash when drawing a styled line
            (default value is arbitrary)
        max_flattening_distance: Max flattening distance in drawing units
            see Path.flattening documentation.
            The backend implementation should calculate an appropriate value,
            like 1 screen- or paper pixel on the output medium, but converted
            into drawing units. Sets Path() approximation accuracy
        circle_approximation_count: Approximate a full circle by `n` segments, arcs
            have proportional less segments. Only used for approximation of arcs
            in banded polylines.
        hatching_timeout: hatching timeout for a single entity, very dense
            hatching patterns can cause a very long execution time, the default
            timeout for a single entity is 30 seconds.
        min_hatch_line_distance: minimum hatch line distance to render, narrower pattern
            lines are rendered as solid filling
        color_policy:
        custom_fg_color: Used for :class:`ColorPolicy.custom` policy, custom foreground
            color as "#RRGGBBAA" color string (RGB+alpha)
        background_policy:
        custom_bg_color: Used for :class:`BackgroundPolicy.custom` policy, custom
            background color as "#RRGGBBAA" color string (RGB+alpha)
        lineweight_policy:
        text_policy:
        image_policy: the method for drawing IMAGE entities

    """

    pdsize: Optional[int] = None  # use $PDSIZE from HEADER section
    pdmode: Optional[int] = None  # use $PDMODE from HEADER section
    measurement: Optional[Measurement] = None
    show_defpoints: bool = False
    proxy_graphic_policy: ProxyGraphicPolicy = ProxyGraphicPolicy.SHOW
    line_policy: LinePolicy = LinePolicy.ACCURATE
    hatch_policy: HatchPolicy = HatchPolicy.NORMAL
    infinite_line_length: float = 20
    lineweight_scaling: float = 1.0
    min_lineweight: Optional[float] = None
    min_dash_length: float = 0.1
    max_flattening_distance: float = disassemble.Primitive.max_flattening_distance
    circle_approximation_count: int = 128
    hatching_timeout: float = 30.0
    # Keep value in sync with ezdxf.render.hatching.MIN_HATCH_LINE_DISTANCE
    min_hatch_line_distance: float = 1e-4
    color_policy: ColorPolicy = ColorPolicy.COLOR
    custom_fg_color: Color = "#000000"
    background_policy: BackgroundPolicy = BackgroundPolicy.DEFAULT
    custom_bg_color: Color = "#ffffff"
    lineweight_policy: LineweightPolicy = LineweightPolicy.ABSOLUTE
    text_policy: TextPolicy = TextPolicy.FILLING
    image_policy: ImagePolicy = ImagePolicy.DISPLAY

    @staticmethod
    def defaults() -> Configuration:
        warnings.warn(
            "use Configuration() instead of Configuration.defaults()",
            DeprecationWarning,
        )
        return Configuration()

    def with_changes(self, **kwargs) -> Configuration:
        """Returns a new frozen :class:`Configuration` object with modified values."""
        params = dataclasses.asdict(self)
        for k, v in kwargs.items():
            params[k] = v
        return Configuration(**params)



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License

from ezdxf.lldxf.const import BYBLOCK, BYLAYER
from ezdxf import colors as clr


def test_decode_by_block():
    assert clr.decode_raw_color(-1056964608) == (
        clr.COLOR_TYPE_BY_BLOCK,
        BYBLOCK,
    )


def test_encode_by_block():
    assert clr.encode_raw_color(BYBLOCK) == -1056964608


def test_decode_by_layer():
    assert clr.decode_raw_color(-1073741824) == (
        clr.COLOR_TYPE_BY_LAYER,
        BYLAYER,
    )


def test_encode_by_layer():
    assert clr.encode_raw_color(BYLAYER) == -1073741824


def test_decode_aci():
    assert clr.decode_raw_color(-1023410164) == (clr.COLOR_TYPE_ACI, 12)


def test_encode_aci():
    assert clr.encode_raw_color(12) == -1023410164


def test_decode_rgb():
    assert clr.decode_raw_color(-1039526882) == (
        clr.COLOR_TYPE_RGB,
        (10, 20, 30),
    )


def test_encode_rgb():
    assert clr.encode_raw_color((10, 20, 30)) == -1039526882



# Copyright (c) 2018-2019, Manfred Moitzi
# License: MIT License
import pytest
import os
import ezdxf
from ezdxf.lldxf.const import versions_supported_by_new


@pytest.fixture(params=versions_supported_by_new)
def drawing(request):
    return ezdxf.new(request.param)


def test_create_anonymous_block(drawing, tmpdir):
    modelspace = drawing.modelspace()
    anonymous_block = drawing.blocks.new_anonymous_block()
    points2d = [
        (0, 0),
        (1, 0),
        (1, 1),
        (0, 1),
        (0, 0),
        (1, 1),
        (0.5, 1.5),
        (0, 1),
        (1, 0),
    ]
    anonymous_block.add_polyline2d(points2d)
    modelspace.add_blockref(anonymous_block.name, (0, 0))
    filename = str(tmpdir.join("anonymous_blocks_%s.dxf" % drawing.dxfversion))
    try:
        drawing.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(
                str(e), drawing.dxfversion
            )
        )
    assert os.path.exists(filename)



# Copyright (c) 2020-2025, Matthew Broadway
# License: MIT License
from __future__ import annotations
from typing import (
    Iterable,
    cast,
    Union,
    Dict,
    Callable,
    Optional,
    TYPE_CHECKING,
)
from typing_extensions import TypeAlias
import contextlib
import math
import os
import pathlib
import logging
import time

import PIL.Image
import PIL.ImageEnhance
import PIL.ImageDraw
import numpy as np


import ezdxf.bbox
from ezdxf.addons.drawing.config import (
    Configuration,
    ProxyGraphicPolicy,
    HatchPolicy,
    ImagePolicy,
)
from ezdxf.addons.drawing.backend import BackendInterface, ImageData
from ezdxf.addons.drawing.properties import (
    RenderContext,
    OLE2FRAME_COLOR,
    Properties,
    Filling,
    LayoutProperties,
    MODEL_SPACE_BG_COLOR,
    PAPER_SPACE_BG_COLOR,
)
from ezdxf.addons.drawing.config import BackgroundPolicy, TextPolicy
from ezdxf.addons.drawing.text import simplified_text_chunks
from ezdxf.addons.drawing.type_hints import FilterFunc
from ezdxf.addons.drawing.gfxproxy import DXFGraphicProxy
from ezdxf.addons.drawing.mtext_complex import complex_mtext_renderer
from ezdxf.entities import (
    DXFEntity,
    DXFGraphic,
    Insert,
    MText,
    Polyline,
    LWPolyline,
    Text,
    Polyface,
    Wipeout,
    Solid,
    Face3d,
    OLE2Frame,
    Point,
    Viewport,
    Image,
)
from ezdxf.tools import clipping_portal
from ezdxf.entities.attrib import BaseAttrib
from ezdxf.entities.polygon import DXFPolygon
from ezdxf.entities.boundary_paths import AbstractBoundaryPath
from ezdxf.entities.textstyle import get_textstyle
from ezdxf.layouts import Layout
from ezdxf.math import Vec2, Vec3, OCS, NULLVEC, Matrix44
from ezdxf.path import (
    Path,
    make_path,
    from_hatch_boundary_path,
    from_vertices,
)
from ezdxf.render import MeshBuilder, TraceBuilder
from ezdxf import reorder
from ezdxf.proxygraphic import ProxyGraphic, ProxyGraphicError
from ezdxf.protocols import SupportsVirtualEntities, virtual_entities
from ezdxf.tools.text import has_inline_formatting_codes, MIN_CAP_HEIGHT
from ezdxf.tools import text_layout
from ezdxf.lldxf import const
from ezdxf.render import hatching
from ezdxf.fonts import fonts
from ezdxf.colors import RGB, RGBA
from ezdxf.npshapes import NumpyPoints2d
from ezdxf import xclip

from .type_hints import Color

if TYPE_CHECKING:
    from .pipeline import AbstractPipeline

__all__ = ["Frontend", "UniversalFrontend"]

TEntityFunc: TypeAlias = Callable[[DXFGraphic, Properties], None]
TDispatchTable: TypeAlias = Dict[str, TEntityFunc]

POST_ISSUE_MSG = (
    "Please post sample DXF file at https://github.com/mozman/ezdxf/issues."
)
logger = logging.getLogger("ezdxf")


class UniversalFrontend:
    """Drawing frontend, responsible for decomposing entities into graphic
    primitives and resolving entity properties. Supports 2D and 3D backends.

    By passing the bounding box cache of the modelspace entities can speed up
    paperspace rendering, because the frontend can filter entities which are not
    visible in the VIEWPORT. Even passing in an empty cache can speed up
    rendering time when multiple viewports need to be processed.

    Args:
        ctx: the properties relevant to rendering derived from a DXF document
        designer: connection between frontend and backend
        config: settings to configure the drawing frontend and backend
        bbox_cache: bounding box cache of the modelspace entities or an empty
            cache which will be filled dynamically when rendering multiple
            viewports or ``None`` to disable bounding box caching at all

    """

    def __init__(
        self,
        ctx: RenderContext,
        pipeline: AbstractPipeline,
        config: Configuration = Configuration(),
        bbox_cache: Optional[ezdxf.bbox.Cache] = None,
    ):
        # RenderContext contains all information to resolve resources for a
        # specific DXF document.
        self.ctx = ctx

        # the render pipeline is the connection between frontend and backend
        self.pipeline = pipeline
        pipeline.set_draw_entities_callback(self.draw_entities_callback)

        # update all configs:
        self.config = ctx.update_configuration(config)
        # backend.configure() is called in pipeline.set_config()
        pipeline.set_config(self.config)

        if self.config.pdsize is None or self.config.pdsize <= 0:
            self.log_message("relative point size is not supported")
            self.config = self.config.with_changes(pdsize=1)

        # Parents entities of current entity/sub-entity
        self.parent_stack: list[DXFGraphic] = []

        self._dispatch = self._build_dispatch_table()

        # Supported DXF entities which use only proxy graphic for rendering:
        self._proxy_graphic_only_entities: set[str] = {
            "MLEADER",  # todo: remove if MLeader.virtual_entities() is implemented
            "MULTILEADER",
            "ACAD_PROXY_ENTITY",
        }

        # Optional bounding box cache, which maybe was created by detecting the
        # modelspace extends. This cache is used when rendering VIEWPORT
        # entities in paperspace to detect if an entity is even visible,
        # this can speed up rendering time if multiple viewports are present.
        # If the bbox_cache is not ``None``, entities not in cache will be
        # added dynamically. This is only beneficial when rendering multiple
        # viewports, as the upfront bounding box calculation adds some rendering
        # time.
        self._bbox_cache = bbox_cache

        # Entity property override function stack:
        self._property_override_functions: list[TEntityFunc] = []
        self.push_property_override_function(self.override_properties)

    @property
    def text_engine(self):
        return self.pipeline.text_engine

    def _build_dispatch_table(self) -> TDispatchTable:
        dispatch_table: TDispatchTable = {
            "POINT": self.draw_point_entity,
            "HATCH": self.draw_hatch_entity,
            "MPOLYGON": self.draw_mpolygon_entity,
            "MESH": self.draw_mesh_entity,
            "WIPEOUT": self.draw_wipeout_entity,
            "MTEXT": self.draw_mtext_entity,
            "OLE2FRAME": self.draw_ole2frame_entity,
            "IMAGE": self.draw_image_entity,
        }
        for dxftype in ("LINE", "XLINE", "RAY"):
            dispatch_table[dxftype] = self.draw_line_entity
        for dxftype in ("TEXT", "ATTRIB", "ATTDEF"):
            dispatch_table[dxftype] = self.draw_text_entity
        for dxftype in ("CIRCLE", "ARC", "ELLIPSE", "SPLINE"):
            dispatch_table[dxftype] = self.draw_curve_entity
        for dxftype in ("3DFACE", "SOLID", "TRACE"):
            dispatch_table[dxftype] = self.draw_solid_entity
        for dxftype in ("POLYLINE", "LWPOLYLINE"):
            dispatch_table[dxftype] = self.draw_polyline_entity
        # Composite entities are detected by implementing the
        # __virtual_entities__() protocol.
        return dispatch_table

    def log_message(self, message: str):
        """Log given message - override to alter behavior."""
        logger.info(message)

    def skip_entity(self, entity: DXFEntity, msg: str) -> None:
        """Called for skipped entities - override to alter behavior."""
        self.log_message(f'skipped entity {str(entity)}. Reason: "{msg}"')

    def exec_property_override(
        self, entity: DXFGraphic, properties: Properties
    ) -> None:
        """Execute entity property override functions. (internal API)"""
        for override_fn in self._property_override_functions:
            override_fn(entity, properties)

    def override_properties(self, entity: DXFGraphic, properties: Properties) -> None:
        """This method can change the resolved properties of an DXF entity.

        The method has access to the DXF entity attributes, the current render context
        and the resolved properties.  It is recommended to modify only the resolved
        `properties` in this method, because the DXF entities are not copies - except
        for virtual entities.

        .. versionchanged:: 1.3.0

            This method is the first function in the stack of new property override
            functions.  It is possible to push additional override functions onto this
            stack, see also :meth:`push_property_override_function`.

        """

    def push_property_override_function(self, override_fn: TEntityFunc) -> None:
        """The override function can change the resolved properties of an DXF entity.

        The override function has access to the DXF entity attributes and the resolved
        properties.  It is recommended to modify only the resolved `properties` in this
        function, because the DXF entities are not copies - except for virtual entities.

        The override functions are called after resolving the DXF attributes of an entity
        and before the :meth:`Frontend.draw_entity` method in the order from first to
        last.

        .. versionadded:: 1.3.0

        """
        self._property_override_functions.append(override_fn)

    def pop_property_override_function(self) -> None:
        """Remove the last function from the property override stack.

        Does not raise an exception if the override stack is empty.

        .. versionadded:: 1.3.0

        """
        if self._property_override_functions:
            self._property_override_functions.pop()

    def draw_layout(
        self,
        layout: Layout,
        finalize: bool = True,
        *,
        filter_func: Optional[FilterFunc] = None,
        layout_properties: Optional[LayoutProperties] = None,
    ) -> None:
        """Draw all entities of the given `layout`.

        Draws the entities of the layout in the default or redefined redraw-order
        and calls the :meth:`finalize` method of the backend if requested.
        The default redraw order is the ascending handle order not the order the
        entities are stored in the layout.

        The method skips invisible entities and entities for which the given
        filter function returns ``False``.

        Args:
            layout: layout to draw of type :class:`~ezdxf.layouts.Layout`
            finalize: ``True`` if the :meth:`finalize` method of the backend
                should be called automatically
            filter_func: function to filter DXf entities, the function should
                return ``False`` if a given entity should be ignored
            layout_properties: override the default layout properties

        """
        if layout_properties is not None:
            self.ctx.current_layout_properties = layout_properties
        else:
            self.ctx.set_current_layout(layout)
        # set background before drawing entities
        self.set_background(self.ctx.current_layout_properties.background_color)
        self.parent_stack = []
        handle_mapping = list(layout.get_redraw_order())
        if handle_mapping:
            self.draw_entities(
                reorder.ascending(layout, handle_mapping),
                filter_func=filter_func,
            )
        else:
            self.draw_entities(
                layout,
                filter_func=filter_func,
            )
        if finalize:
            self.pipeline.finalize()

    def set_background(self, color: Color) -> None:
        policy = self.config.background_policy
        override = True
        if policy == BackgroundPolicy.DEFAULT:
            override = False
        elif policy == BackgroundPolicy.OFF:
            color = "#ffffff00"  # white, fully transparent
        elif policy == BackgroundPolicy.WHITE:
            color = "#ffffff"
        elif policy == BackgroundPolicy.BLACK:
            color = "#000000"
        elif policy == BackgroundPolicy.PAPERSPACE:
            color = PAPER_SPACE_BG_COLOR
        elif policy == BackgroundPolicy.MODELSPACE:
            color = MODEL_SPACE_BG_COLOR
        elif policy == BackgroundPolicy.CUSTOM:
            color = self.config.custom_bg_color
        if override:
            self.ctx.current_layout_properties.set_colors(color)
        self.pipeline.set_background(color)

    def draw_entities(
        self,
        entities: Iterable[DXFGraphic],
        *,
        filter_func: Optional[FilterFunc] = None,
    ) -> None:
        """Draw the given `entities`. The method skips invisible entities
        and entities for which the given filter function returns ``False``.

        """
        _draw_entities(self, self.ctx, entities, filter_func=filter_func)

    def draw_entities_callback(
        self, ctx: RenderContext, entities: Iterable[DXFGraphic]
    ) -> None:
        _draw_entities(self, ctx, entities)

    def draw_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        """Draw a single DXF entity.

        Args:
            entity: DXF entity inherited from DXFGraphic()
            properties: resolved entity properties

        """
        self.pipeline.enter_entity(entity, properties)
        if not entity.is_virtual:
            # top level entity
            self.pipeline.set_current_entity_handle(entity.dxf.handle)
        if (
            entity.proxy_graphic
            and self.config.proxy_graphic_policy == ProxyGraphicPolicy.PREFER
        ):
            self.draw_proxy_graphic(entity.proxy_graphic, entity.doc)
        else:
            draw_method = self._dispatch.get(entity.dxftype(), None)
            if draw_method is not None:
                draw_method(entity, properties)
            # Composite entities (INSERT, DIMENSION, ...) have to implement the
            # __virtual_entities__() protocol.
            # Unsupported DXF types which have proxy graphic, are wrapped into
            # DXFGraphicProxy, which also implements the __virtual_entities__()
            # protocol.
            elif isinstance(entity, SupportsVirtualEntities):
                assert isinstance(entity, DXFGraphic)
                # The __virtual_entities__() protocol does not distinguish
                # content from DXF entities or from proxy graphic.
                # In the long run ACAD_PROXY_ENTITY should be the only
                # supported DXF entity which uses proxy graphic. Unsupported
                # DXF entities (DXFGraphicProxy) do not get to this point if
                # proxy graphic is ignored.
                if (
                    self.config.proxy_graphic_policy != ProxyGraphicPolicy.IGNORE
                    or entity.dxftype() not in self._proxy_graphic_only_entities
                ):
                    self.draw_composite_entity(entity, properties)
            else:
                self.skip_entity(entity, "unsupported")

        self.pipeline.exit_entity(entity)

    def draw_line_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        d, dxftype = entity.dxf, entity.dxftype()
        if dxftype == "LINE":
            self.pipeline.draw_line(d.start, d.end, properties)

        elif dxftype in ("XLINE", "RAY"):
            start = d.start
            delta = d.unit_vector * self.config.infinite_line_length
            if dxftype == "XLINE":
                self.pipeline.draw_line(
                    start - delta / 2, start + delta / 2, properties
                )
            elif dxftype == "RAY":
                self.pipeline.draw_line(start, start + delta, properties)
        else:
            raise TypeError(dxftype)

    def draw_text_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        if self.config.text_policy == TextPolicy.IGNORE:
            return
        # fixed text height set by associated text-style is ignored
        if entity.dxf.height < MIN_CAP_HEIGHT:
            self.skip_entity(entity, "text height too small")
            return
        # Draw embedded MTEXT entity as virtual MTEXT entity:
        if isinstance(entity, BaseAttrib) and entity.has_embedded_mtext_entity:
            self.draw_mtext_entity(entity.virtual_mtext_entity(), properties)
        elif is_spatial_text(Vec3(entity.dxf.extrusion)):
            self.draw_text_entity_3d(entity, properties)
        else:
            self.draw_text_entity_2d(entity, properties)

    def get_font_face(self, properties: Properties) -> fonts.FontFace:
        font_face = properties.font
        if font_face is None:
            return self.pipeline.default_font_face
        return font_face

    def draw_text_entity_2d(self, entity: DXFGraphic, properties: Properties) -> None:
        if isinstance(entity, Text):
            for line, transform, cap_height in simplified_text_chunks(
                entity, self.text_engine, font_face=self.get_font_face(properties)
            ):
                self.pipeline.draw_text(
                    line, transform, properties, cap_height, entity.dxftype()
                )
        else:
            raise TypeError(entity.dxftype())

    def draw_text_entity_3d(self, entity: DXFGraphic, properties: Properties) -> None:
        self.skip_entity(entity, "3D text not supported")

    def draw_mtext_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        # fixed text height set by associated text-style is ignored
        if self.config.text_policy == TextPolicy.IGNORE:
            return
        # Do not skip MTEXT entities with char height 0.
        # Text height is maybe changed by inline-codes.
        mtext = cast(MText, entity)
        if is_spatial_text(Vec3(mtext.dxf.extrusion)):
            self.skip_entity(mtext, "3D MTEXT not supported")
            return
        if is_complex_mtext(mtext):
            try:
                self.draw_complex_mtext(mtext, properties)
            except text_layout.LayoutError as e:
                print(f"skipping {str(mtext)} - {str(e)}")
        else:
            self.draw_simple_mtext(mtext, properties)

    def draw_simple_mtext(self, mtext: MText, properties: Properties) -> None:
        """Draw the content of a MTEXT entity without inline formatting codes."""
        for line, transform, cap_height in simplified_text_chunks(
            mtext, self.text_engine, font_face=self.get_font_face(properties)
        ):
            self.pipeline.draw_text(
                line, transform, properties, cap_height, mtext.dxftype()
            )

    def draw_complex_mtext(self, mtext: MText, properties: Properties) -> None:
        """Draw the content of a MTEXT entity including inline formatting codes."""
        complex_mtext_renderer(self.ctx, self.pipeline, mtext, properties)

    def draw_curve_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        try:
            path = make_path(entity)
        except AttributeError:  # API usage error
            raise TypeError(f"Unsupported DXF type {entity.dxftype()}")
        self.pipeline.draw_path(path, properties)

    def draw_point_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        point = cast(Point, entity)
        pdmode = self.config.pdmode
        pdsize = self.config.pdsize
        assert pdmode is not None
        assert pdsize is not None

        # Defpoints are regular POINT entities located at the "defpoints" layer:
        if properties.layer.lower() == "defpoints":
            if not self.config.show_defpoints:
                return
            else:  # Render defpoints as dimensionless points:
                pdmode = 0

        if pdmode == 0:
            self.pipeline.draw_point(entity.dxf.location, properties)
        else:
            for entity in point.virtual_entities(pdsize, pdmode):
                dxftype = entity.dxftype()
                if dxftype == "LINE":
                    start = entity.dxf.start
                    end = entity.dxf.end
                    if start.isclose(end):
                        self.pipeline.draw_point(start, properties)
                    else:  # direct draw by backend is OK!
                        self.pipeline.draw_line(start, end, properties)
                    pass
                elif dxftype == "CIRCLE":
                    self.draw_curve_entity(entity, properties)
                else:
                    raise ValueError(dxftype)

    def draw_solid_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        if isinstance(entity, Face3d):
            dxf = entity.dxf
            try:
                # this implementation supports all features of example file:
                # examples_dxf/3dface.dxf without changing the behavior of
                # Face3d.wcs_vertices() which removes the last vertex if
                # duplicated.
                points = [dxf.vtx0, dxf.vtx1, dxf.vtx2, dxf.vtx3, dxf.vtx0]
            except AttributeError:
                # all 4 vertices are required, otherwise the entity is invalid
                # for AutoCAD
                self.skip_entity(entity, "missing required vertex attribute")
                return
            edge_visibility = entity.get_edges_visibility()
            if all(edge_visibility):
                self.pipeline.draw_path(from_vertices(points), properties)
            else:
                for a, b, visible in zip(points, points[1:], edge_visibility):
                    if visible:
                        self.pipeline.draw_line(a, b, properties)

        elif isinstance(entity, Solid):
            # set solid fill type for SOLID and TRACE
            properties.filling = Filling()
            self.pipeline.draw_filled_polygon(
                entity.wcs_vertices(close=False), properties
            )
        else:
            raise TypeError("API error, requires a SOLID, TRACE or 3DFACE entity")

    def draw_hatch_pattern(
        self, polygon: DXFPolygon, paths: list[Path], properties: Properties
    ):
        if polygon.pattern is None or len(polygon.pattern.lines) == 0:
            return
        ocs = polygon.ocs()
        elevation = polygon.dxf.elevation.z
        properties.linetype_pattern = tuple()
        lines: list[tuple[Vec3, Vec3]] = []

        t0 = time.perf_counter()
        max_time = self.config.hatching_timeout

        def timeout() -> bool:
            if time.perf_counter() - t0 > max_time:
                print(
                    f"hatching timeout of {max_time}s reached for {str(polygon)} - aborting"
                )
                return True
            return False

        for baseline in hatching.pattern_baselines(
            polygon,
            min_hatch_line_distance=self.config.min_hatch_line_distance,
            jiggle_origin=True,
        ):
            for line in hatching.hatch_paths(baseline, paths, timeout):
                line_pattern = baseline.pattern_renderer(line.distance)
                for s, e in line_pattern.render(line.start, line.end):
                    if ocs.transform:
                        s, e = (
                            ocs.to_wcs((s.x, s.y, elevation)),
                            ocs.to_wcs((e.x, e.y, elevation)),
                        )
                    lines.append((s, e))
        self.pipeline.draw_solid_lines(lines, properties)

    def draw_hatch_entity(
        self,
        entity: DXFGraphic,
        properties: Properties,
        *,
        loops: Optional[list[Path]] = None,
    ) -> None:
        if properties.filling is None:
            return
        filling = properties.filling
        show_only_outline = False
        hatch_policy = self.config.hatch_policy
        if hatch_policy == HatchPolicy.NORMAL:
            pass
        elif hatch_policy == HatchPolicy.IGNORE:
            return
        elif hatch_policy == HatchPolicy.SHOW_SOLID:
            filling = Filling()  # solid filling
        elif hatch_policy == HatchPolicy.SHOW_OUTLINE:
            filling = Filling()  # solid filling
            show_only_outline = True

        polygon = cast(DXFPolygon, entity)
        if filling.type == Filling.PATTERN:
            if loops is None:
                loops = hatching.hatch_boundary_paths(polygon, filter_text_boxes=True)
            try:
                self.draw_hatch_pattern(polygon, loops, properties)
            except hatching.DenseHatchingLinesError:
                pass  # fallthrough to solid fill rendering
            else:
                return

        # draw SOLID filling
        ocs = polygon.ocs()
        # all OCS coordinates have the same z-axis stored as vector (0, 0, z),
        # default (0, 0, 0)
        elevation = entity.dxf.elevation.z
        paths: list[Path]
        if loops is not None:  # only MPOLYGON
            paths = loops
        else:  # only HATCH
            boundary_paths = list(
                polygon.paths.rendering_paths(polygon.dxf.hatch_style)
            )
            paths = closed_loops(boundary_paths, ocs, elevation)
        if show_only_outline:
            for p in ignore_text_boxes(paths):
                self.pipeline.draw_path(p, properties)
            return
        if paths:
            self.pipeline.draw_filled_paths(ignore_text_boxes(paths), properties)

    def draw_mpolygon_entity(self, entity: DXFGraphic, properties: Properties):
        def resolve_fill_color() -> str:
            return self.ctx.resolve_aci_color(entity.dxf.fill_color, properties.layer)

        polygon = cast(DXFPolygon, entity)
        ocs = polygon.ocs()
        elevation: float = polygon.dxf.elevation.z
        offset = Vec3(polygon.dxf.get("offset_vector", NULLVEC))
        # MPOLYGON does not support hatch styles, all paths are rendered.
        loops = closed_loops(polygon.paths, ocs, elevation, offset)  # type: ignore

        line_color: str = properties.color
        assert properties.filling is not None
        pattern_name: str = properties.filling.name  # normalized pattern name
        # 1. draw filling
        if polygon.dxf.solid_fill:
            properties.filling.type = Filling.SOLID
            if polygon.gradient is not None and polygon.gradient.number_of_colors > 0:
                # true color filling is stored as gradient
                properties.color = str(properties.filling.gradient_color1)
            else:
                properties.color = resolve_fill_color()
            self.draw_hatch_entity(entity, properties, loops=loops)

        # checking properties.filling.type == Filling.PATTERN is not sufficient:
        elif pattern_name and pattern_name != "SOLID":
            # line color is also pattern color: properties.color
            self.draw_hatch_entity(entity, properties, loops=loops)

        # 2. draw boundary paths
        properties.color = line_color
        # draw boundary paths as lines
        for loop in loops:
            self.pipeline.draw_path(loop, properties)

    def draw_wipeout_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        wipeout = cast(Wipeout, entity)
        properties.filling = Filling()
        properties.color = self.ctx.current_layout_properties.background_color
        clipping_polygon = wipeout.boundary_path_wcs()
        self.pipeline.draw_filled_polygon(clipping_polygon, properties)

    def draw_viewport(self, vp: Viewport) -> None:
        # the "active" viewport and invisible viewports should be filtered at this
        # stage, see function _draw_viewports()
        if vp.dxf.status < 1:
            return

        if not vp.is_top_view:
            self.log_message("Cannot render non top-view viewports")
            return
        self.pipeline.draw_viewport(vp, self.ctx, self._bbox_cache)

    def draw_ole2frame_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        ole2frame = cast(OLE2Frame, entity)
        bbox = ole2frame.bbox()
        if not bbox.is_empty:
            self._draw_filled_rect(bbox.rect_vertices(), OLE2FRAME_COLOR)

    def draw_image_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        image = cast(Image, entity)
        image_policy = self.config.image_policy
        image_def = image.image_def
        assert image_def is not None

        if image_policy in (
            ImagePolicy.DISPLAY,
            ImagePolicy.RECT,
            ImagePolicy.MISSING,
        ):
            loaded_image = None
            show_filename_if_missing = True

            if (
                image_policy == ImagePolicy.RECT
                or not image.dxf.flags & Image.SHOW_IMAGE
            ):
                loaded_image = None
                show_filename_if_missing = False
            elif (
                image_policy != ImagePolicy.MISSING
                and self.ctx.document_dir is not None
            ):
                image_path = _find_image_path(
                    self.ctx.document_dir, image_def.dxf.filename
                )
                with contextlib.suppress(FileNotFoundError):
                    loaded_image = PIL.Image.open(image_path)

            if loaded_image is not None:
                color: RGB | RGBA
                loaded_image = loaded_image.convert("RGBA")  # type: ignore

                if image.dxf.contrast != 50:
                    # note: this is only an approximation.
                    # Unclear what the exact operation AutoCAD uses
                    amount = image.dxf.contrast / 50
                    loaded_image = PIL.ImageEnhance.Contrast(loaded_image).enhance(  # type: ignore
                        amount
                    )

                if image.dxf.fade != 0:
                    # note: this is only an approximation.
                    # Unclear what the exact operation AutoCAD uses
                    amount = image.dxf.fade / 100
                    color = RGB.from_hex(
                        self.ctx.current_layout_properties.background_color
                    )
                    loaded_image = _blend_image_towards(loaded_image, amount, color)  # type: ignore
                if image.dxf.brightness != 50:
                    # note: this is only an approximation.
                    # Unclear what the exact operation AutoCAD uses
                    amount = image.dxf.brightness / 50 - 1
                    if amount > 0:
                        color = RGBA(255, 255, 255, 255)
                    else:
                        color = RGBA(0, 0, 0, 255)
                        amount = -amount
                    loaded_image = _blend_image_towards(loaded_image, amount, color)  # type: ignore

                if not image.dxf.flags & Image.USE_TRANSPARENCY:
                    loaded_image.putalpha(255)  # type: ignore

                if image.transparency != 0.0:
                    loaded_image = _multiply_alpha(
                        loaded_image,  # type: ignore
                        1.0 - image.transparency,
                    )
                image_data = ImageData(
                    image=np.array(loaded_image),
                    transform=image.get_wcs_transform(),
                    pixel_boundary_path=NumpyPoints2d(image.pixel_boundary_path()),
                    use_clipping_boundary=image.dxf.flags & Image.USE_CLIPPING_BOUNDARY,
                    remove_outside=image.dxf.clip_mode == 0,
                )
                self.pipeline.draw_image(image_data, properties)

            elif show_filename_if_missing:
                default_cap_height = 20
                text = image_def.dxf.filename
                if not text.strip():
                    text = "<no filename>"
                font = self.pipeline.text_engine.get_font(
                    self.get_font_face(properties)
                )
                text_width = font.text_width_ex(text, default_cap_height)
                image_size = image.dxf.image_size
                desired_width = image_size.x * 0.75
                scale = desired_width / text_width
                translate = Matrix44.translate(
                    (image_size.x - desired_width) / 2,
                    (image_size.y - default_cap_height * scale) / 2,
                    0,
                )
                transform = (
                    Matrix44.scale(scale) @ translate @ image.get_wcs_transform()
                )
                self.pipeline.draw_text(
                    text,
                    transform,
                    properties,
                    default_cap_height,
                )

            points = [v.vec2 for v in image.boundary_path_wcs()]
            self.pipeline.draw_solid_lines(list(zip(points, points[1:])), properties)

        elif self.config.image_policy == ImagePolicy.PROXY:
            self.draw_proxy_graphic(entity.proxy_graphic, entity.doc)

        elif self.config.image_policy == ImagePolicy.IGNORE:
            pass

        else:
            raise ValueError(self.config.image_policy)

    def _draw_filled_rect(
        self,
        points: Iterable[Vec2],
        color: str,
    ) -> None:
        props = Properties()
        props.color = color
        # default SOLID filling
        props.filling = Filling()
        self.pipeline.draw_filled_polygon(points, props)

    def draw_mesh_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        builder = MeshBuilder.from_mesh(entity)  # type: ignore
        self.draw_mesh_builder_entity(builder, properties)

    def draw_mesh_builder_entity(
        self, builder: MeshBuilder, properties: Properties
    ) -> None:
        for face in builder.faces_as_vertices():
            self.pipeline.draw_path(
                from_vertices(face, close=True), properties=properties
            )

    def draw_polyline_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        dxftype = entity.dxftype()
        if dxftype == "POLYLINE":
            e = cast(Polyface, entity)
            if e.is_polygon_mesh or e.is_poly_face_mesh:
                # draw 3D mesh or poly-face entity
                self.draw_mesh_builder_entity(
                    MeshBuilder.from_polyface(e),
                    properties,
                )
                return

        entity = cast(Union[LWPolyline, Polyline], entity)
        is_lwpolyline = dxftype == "LWPOLYLINE"

        if entity.has_width:  # draw banded 2D polyline
            elevation = 0.0
            ocs = entity.ocs()
            transform = ocs.transform
            if transform:
                if is_lwpolyline:  # stored as float
                    elevation = entity.dxf.elevation
                else:  # stored as vector (0, 0, elevation)
                    elevation = Vec3(entity.dxf.elevation).z

            trace = TraceBuilder.from_polyline(
                entity, segments=self.config.circle_approximation_count // 2
            )
            for polygon in trace.polygons():  # polygon is a sequence of Vec2()
                if len(polygon) < 3:
                    continue
                if transform:
                    points = ocs.points_to_wcs(
                        Vec3(v.x, v.y, elevation) for v in polygon
                    )
                else:
                    points = polygon  # type: ignore
                # Set default SOLID filling for LWPOLYLINE
                properties.filling = Filling()
                self.pipeline.draw_filled_polygon(points, properties)
            return
        polyline_path = make_path(entity)
        if len(polyline_path):
            self.pipeline.draw_path(polyline_path, properties)

    def draw_composite_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        def draw_insert(insert: Insert):
            clip = xclip.XClip(insert)
            is_clipping_active = clip.has_clipping_path and clip.is_clipping_enabled

            if is_clipping_active:
                boundary_path = clip.get_wcs_clipping_path()
                if not boundary_path.is_inverted_clip:
                    clipping_shape = clipping_portal.find_best_clipping_shape(
                        boundary_path.vertices
                    )
                else:  # inverted clipping path
                    clipping_shape = clipping_portal.make_inverted_clipping_shape(
                        boundary_path.inner_polygon(),
                        outer_bounds=boundary_path.outer_bounds(),
                    )
                self.pipeline.push_clipping_shape(clipping_shape, None)

            # draw_entities() includes the visibility check:
            self.draw_entities(
                insert.virtual_entities(
                    skipped_entity_callback=self.skip_entity
                    # TODO: redraw_order=True?
                )
            )

            if is_clipping_active and clip.get_xclip_frame_policy():
                self.pipeline.draw_path(
                    path=from_vertices(boundary_path.inner_polygon(), close=True),
                    properties=properties,
                )
                self.pipeline.pop_clipping_shape()

            # Draw ATTRIB entities at last, see #1321
            # Block reference attributes are located __outside__ the block reference!
            self.draw_entities(insert.attribs)

        if isinstance(entity, Insert):
            self.ctx.push_state(properties)
            if entity.mcount > 1:
                for virtual_insert in entity.multi_insert():
                    draw_insert(virtual_insert)
            else:
                draw_insert(entity)
            self.ctx.pop_state()
        elif isinstance(entity, SupportsVirtualEntities):
            # draw_entities() includes the visibility check:
            try:
                self.draw_entities(virtual_entities(entity))
            except ProxyGraphicError as e:
                print(str(e))
                print(POST_ISSUE_MSG)
        else:
            raise TypeError(entity.dxftype())

    def draw_proxy_graphic(self, data: bytes | None, doc) -> None:
        if not data:
            return
        try:
            self.draw_entities(virtual_entities(ProxyGraphic(data, doc)))
        except ProxyGraphicError as e:
            print(str(e))
            print(POST_ISSUE_MSG)


class Frontend(UniversalFrontend):
    """Drawing frontend for 2D backends, responsible for decomposing entities into
    graphic primitives and resolving entity properties.

    By passing the bounding box cache of the modelspace entities can speed up
    paperspace rendering, because the frontend can filter entities which are not
    visible in the VIEWPORT. Even passing in an empty cache can speed up
    rendering time when multiple viewports need to be processed.

    Args:
        ctx: the properties relevant to rendering derived from a DXF document
        out: the 2D backend to draw to
        config: settings to configure the drawing frontend and backend
        bbox_cache: bounding box cache of the modelspace entities or an empty
            cache which will be filled dynamically when rendering multiple
            viewports or ``None`` to disable bounding box caching at all

    """

    def __init__(
        self,
        ctx: RenderContext,
        out: BackendInterface,
        config: Configuration = Configuration(),
        bbox_cache: Optional[ezdxf.bbox.Cache] = None,
    ):
        from .pipeline import RenderPipeline2d

        super().__init__(ctx, RenderPipeline2d(out), config, bbox_cache)


def is_spatial_text(extrusion: Vec3) -> bool:
    # note: the magnitude of the extrusion vector has no effect on text scale
    return not math.isclose(extrusion.x, 0) or not math.isclose(extrusion.y, 0)


def closed_loops(
    paths: list[AbstractBoundaryPath],
    ocs: OCS,
    elevation: float,
    offset: Vec3 = NULLVEC,
) -> list[Path]:
    loops = []
    for boundary in paths:
        path = from_hatch_boundary_path(boundary, ocs, elevation, offset)
        assert isinstance(
            path.user_data, const.BoundaryPathState
        ), "missing attached boundary path state"
        for sub_path in path.sub_paths():
            if len(sub_path):
                sub_path.close()
                loops.append(sub_path)
    return loops


def ignore_text_boxes(paths: Iterable[Path]) -> Iterable[Path]:
    """Filters text boxes from paths if BoundaryPathState() information is
    attached.
    """
    for path in paths:
        if (
            isinstance(path.user_data, const.BoundaryPathState)
            and path.user_data.textbox
        ):
            continue  # skip text box paths
        yield path


def _draw_entities(
    frontend: UniversalFrontend,
    ctx: RenderContext,
    entities: Iterable[DXFGraphic],
    *,
    filter_func: Optional[FilterFunc] = None,
) -> None:
    if filter_func is not None:
        entities = filter(filter_func, entities)
    viewports: list[Viewport] = []
    for entity in entities:
        if isinstance(entity, Viewport):
            viewports.append(entity)
            continue
        if not isinstance(entity, DXFGraphic):
            if frontend.config.proxy_graphic_policy != ProxyGraphicPolicy.IGNORE:
                entity = DXFGraphicProxy(entity)
            else:
                frontend.skip_entity(entity, "Cannot parse DXF entity")
                continue
        properties = ctx.resolve_all(entity)
        frontend.exec_property_override(entity, properties)
        if properties.is_visible:
            frontend.draw_entity(entity, properties)
        else:
            frontend.skip_entity(entity, "invisible")
    _draw_viewports(frontend, viewports)


def _draw_viewports(frontend: UniversalFrontend, viewports: list[Viewport]) -> None:
    # The VIEWPORT attributes "id" and "status" are very unreliable, maybe because of
    # the "great" documentation by Autodesk.
    # Viewport status field: (according to the DXF Reference)
    # -1 = On, but is fully off-screen, or is one of the viewports that is not
    #      active because the $MAXACTVP count is currently being exceeded.
    #  0 = Off
    # <positive value> = On and active. The value indicates the order of
    # stacking for the viewports, where 1 is the "active" viewport, 2 is the
    # next, and so on.
    viewports.sort(key=lambda e: e.dxf.status)
    # Remove all invisible viewports:
    viewports = [vp for vp in viewports if vp.dxf.status > 0]
    if not viewports:
        return
    # The "active" viewport determines how the paperspace layout is presented as a
    # whole (location & zoom state).
    # Maybe there are more than one "active" viewports, just remove the first one,
    # or there is no "active" viewport at all - in this case the "status" attribute
    # is not reliable at all - but what else is there to do?  The "active" layout should
    # have the id "1", but this information is also not reliable.
    if viewports[0].dxf.get("status", 1) == 1:
        viewports.pop(0)
    # Draw viewports in order of "status"
    for viewport in viewports:
        frontend.draw_viewport(viewport)


def _blend_image_towards(
    image: PIL.Image.Image, amount: float, color: RGB | RGBA
) -> PIL.Image.Image:
    original_alpha = image.split()[-1]
    destination = PIL.Image.new("RGBA", image.size, color)
    updated_image = PIL.Image.blend(image, destination, amount)
    updated_image.putalpha(original_alpha)
    return updated_image


def _multiply_alpha(image: PIL.Image.Image, amount: float) -> PIL.Image.Image:
    output_image = np.array(image, dtype=np.float64)
    output_image[:, :, 3] *= amount
    return PIL.Image.fromarray(output_image.astype(np.uint8), "RGBA")


def _find_image_path(document_dir: pathlib.Path, filename: str) -> pathlib.Path:
    # BricsCAD stores the path to the image as full-path or relative-path if so
    # set in the "Attach Raster Image" dialog.
    # See notes in knowledge graph: [[IMAGE File Paths]]
    # https://ezdxf.mozman.at/notes/#/page/image%20file%20paths

    # BricsCAD/AutoCAD stores filepaths with backslashes.
    filename = filename.replace("\\", os.path.sep)

    # try absolute path:
    filepath = pathlib.Path(filename)
    if filepath.exists():
        return filepath

    # try relative path to document:
    filepath = document_dir / filename
    filepath = filepath.resolve()
    if filepath.exists():
        return filepath

    # try document dir:
    filepath = document_dir / pathlib.Path(filename).name  # stem + suffix
    return filepath


def is_complex_mtext(mtext: MText) -> bool:
    if mtext.has_columns:
        return True
    if has_inline_formatting_codes(mtext.text):
        return True
    if mtext.doc is None:  # cannot determine textstyle
        return False
    textstyle = get_textstyle(mtext)
    return textstyle.dxf.width != 1 or textstyle.dxf.oblique != 0



# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
from binascii import unhexlify
from ezdxf.lldxf.types import DXFBinaryTag
from ezdxf.lldxf.tags import binary_data_to_dxf_tags


def test_init():
    tag = DXFBinaryTag.from_string(310, "FFFF")
    assert tag == (310, b"\xff\xff")

    tag = DXFBinaryTag(310, b"\xff\xff")
    assert tag == (310, b"\xff\xff")

    tag2 = DXFBinaryTag.from_string(code=310, value="FFFF")
    assert tag2 == (310, b"\xff\xff")


def test_index_able():
    tag = DXFBinaryTag.from_string(310, "FFFF")
    assert tag[0] == 310
    assert tag[1] == b"\xff\xff"


def test_dxf_str():
    assert DXFBinaryTag.from_string(310, "FFFF").tostring() == "FFFF"
    assert DXFBinaryTag.from_string(310, "FFFF").dxfstr() == "310\nFFFF\n"


def test_long_string():
    tag = DXFBinaryTag.from_string(
        310,
        "414349532042696E61727946696C652855000000000000020000000C00000007104175746F6465736B204175746F434144071841534D203231392E302E302E3536303020556E6B6E6F776E071853756E204D61792020342031353A34373A3233203230313406000000000000F03F068DEDB5A0F7C6B03E06BBBDD7D9DF7CDB",
    )
    assert len(tag.value) == 127
    clone = tag.clone()
    assert tag.value == clone.value


def test_hexstr_to_bytes():
    import array

    s = "".join("%0.2X" % byte for byte in range(256))
    assert len(s) == 512
    bytes_ = unhexlify(s)
    assert len(bytes_) == 256
    for x, y in zip(array.array("B", bytes_), range(256)):
        assert x == y
    tag = DXFBinaryTag(310, bytes_)
    assert tag.tostring() == s


class TestBinaryDataToDXFTags:
    def test_for_preceding_length_tag(self):
        tags = binary_data_to_dxf_tags(b"0123456789")
        assert tags[0] == (160, 10)

    def test_if_data_tag_values_are_bytes(self):
        tags = binary_data_to_dxf_tags(b"0123456789")
        assert isinstance(tags[1].value, bytes)

    def test_if_empty_data_creates_a_length_tag(self):
        tags = binary_data_to_dxf_tags(b"")
        assert tags[0] == (160, 0)

    def test_if_tag_value_size_is_limited_to_chunk_size(self):
        data = b"0123456789" * 10
        tags = binary_data_to_dxf_tags(data, value_size=10)
        assert len(tags) == 11
        assert tags[0] == (160, 100)
        assert all(len(t.value) == 10 for t in tags[1:])

    def test_if_merged_tag_values_matches_source_data(self):
        data = b"0123456789" * 10
        tags = binary_data_to_dxf_tags(data, value_size=10)
        content = b"".join(t.value for t in tags[1:])
        assert data == content



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Sequence, NamedTuple, Any, Iterator
from typing_extensions import Self
import abc
import copy
import enum
import math
from .deps import (
    Vec2,
    Path,
    colors,
    Matrix44,
    BoundingBox2d,
)
from .properties import Properties, Pen
from ezdxf.npshapes import NumpyPath2d, NumpyPoints2d

# Page coordinates are always plot units:
# 1 plot unit (plu) = 0.025mm
# 40 plu = 1mm
# 1016 plu = 1 inch
# 3.39 plu = 1 dot @300 dpi
# positive x-axis is horizontal from left to right
# positive y-axis is vertical from bottom to top


class Backend(abc.ABC):
    """Abstract base class for implementing a low level output backends."""

    @abc.abstractmethod
    def draw_polyline(self, properties: Properties, points: Sequence[Vec2]) -> None:
        """Draws a polyline from a sequence `points`. The input coordinates are page
        coordinates in plot units. The `points` sequence can contain 0 or more
        points!

        Args:
            properties: display :class:`Properties` for the polyline
            points: sequence of :class:`ezdxf.math.Vec2` instances

        """
        ...

    @abc.abstractmethod
    def draw_paths(
        self, properties: Properties, paths: Sequence[Path], filled: bool
    ) -> None:
        """Draws filled or outline paths from the sequence of `paths`. The input coordinates
        are page coordinates in plot units. The `paths` sequence can contain 0 or more
        single :class:`~ezdxf.path.Path` instances. Draws outline paths if
        Properties.FillType is NONE and filled paths otherwise.

        Args:
            properties: display :class:`Properties` for the filled polygon
            paths: sequence of single :class:`ezdxf.path.Path` instances
            filled: draw filled paths if ``True`` otherwise outline paths

        """
        ...


class RecordType(enum.Enum):
    POLYLINE = enum.auto()
    FILLED_PATHS = enum.auto()
    OUTLINE_PATHS = enum.auto()


class DataRecord(NamedTuple):
    type: RecordType
    property_hash: int
    data: Any


class Recorder(Backend):
    """The :class:`Recorder` class records the output of the :class:`Plotter` class.

    All input coordinates are page coordinates:

    - 1 plot unit (plu) = 0.025mm
    - 40 plu = 1 mm
    - 1016 plu = 1 inch

    """

    def __init__(self) -> None:
        self._records: list[DataRecord] = []
        self._properties: dict[int, Properties] = {}
        self._pens: Sequence[Pen] = []

    def player(self) -> Player:
        """Returns a :class:`Player` instance with the original recordings. Make a copy
        of this player to protect the original recordings from being modified::

            safe_player = recorder.player().copy()

        """
        return Player(self._records, self._properties)

    def draw_polyline(self, properties: Properties, points: Sequence[Vec2]) -> None:
        self.store(RecordType.POLYLINE, properties, NumpyPoints2d(points))

    def draw_paths(
        self, properties: Properties, paths: Sequence[Path], filled: bool
    ) -> None:
        data = tuple(map(NumpyPath2d, paths))
        record_type = RecordType.FILLED_PATHS if filled else RecordType.OUTLINE_PATHS
        self.store(record_type, properties, data)

    def store(self, record_type: RecordType, properties: Properties, args) -> None:
        prop_hash = properties.hash()
        if prop_hash not in self._properties:
            self._properties[prop_hash] = properties.copy()
        self._records.append(DataRecord(record_type, prop_hash, args))
        if len(self._pens) != len(properties.pen_table):
            self._pens = list(properties.pen_table.values())


class Player:
    """This class replays the recordings of the :class:`Recorder` class on another
    backend. The class can modify the recorded output.
    """

    def __init__(self, records: list[DataRecord], properties: dict[int, Properties]):
        self._records: list[DataRecord] = records
        self._properties: dict[int, Properties] = properties
        self._bbox = BoundingBox2d()

    def __copy__(self) -> Self:
        """Returns a new :class:`Player` instance with a copy of recordings."""
        records = copy.deepcopy(self._records)
        player = self.__class__(records, self._properties)
        player._bbox = self._bbox.copy()
        return player

    copy = __copy__

    def recordings(self) -> Iterator[tuple[RecordType, Properties, Any]]:
        """Yields all recordings as `(RecordType, Properties, Data)` tuples.

        The content of the `Data` field is determined by the enum :class:`RecordType`:

        - :attr:`RecordType.POLYLINE` returns a :class:`NumpyPoints2d` instance
        - :attr:`RecordType.FILLED_POLYGON` returns a tuple of :class:`NumpyPath2d` instances

        """
        props = self._properties
        for record in self._records:
            yield record.type, props[record.property_hash], record.data

    def bbox(self) -> BoundingBox2d:
        """Returns the bounding box of all recorded polylines and polygons as
        :class:`~ezdxf.math.BoundingBox2d`.
        """
        if not self._bbox.has_data:
            self.update_bbox()
        return self._bbox

    def update_bbox(self) -> None:
        points: list[Vec2] = []
        for record in self._records:
            if record.type == RecordType.POLYLINE:
                points.extend(record.data.extents())
            else:
                for path in record.data:
                    points.extend(path.extents())
        self._bbox = BoundingBox2d(points)

    def replay(self, backend: Backend) -> None:
        """Replay the recording on another backend."""
        current_props = Properties()
        props = self._properties
        for record in self._records:
            current_props = props.get(record.property_hash, current_props)
            if record.type == RecordType.POLYLINE:
                backend.draw_polyline(current_props, record.data.vertices())
            else:
                paths = [p.to_path2d() for p in record.data]
                backend.draw_paths(
                    current_props, paths, filled=record.type == RecordType.FILLED_PATHS
                )

    def transform(self, m: Matrix44) -> None:
        """Transforms the recordings by a transformation matrix `m` of type
        :class:`~ezdxf.math.Matrix44`.
        """
        for record in self._records:
            if record.type == RecordType.POLYLINE:
                record.data.transform_inplace(m)
            else:
                for path in record.data:
                    path.transform_inplace(m)

        if self._bbox.has_data:
            # fast, but maybe inaccurate update
            self._bbox = BoundingBox2d(m.fast_2d_transform(self._bbox.rect_vertices()))

    def sort_filled_paths(self) -> None:
        """Sort filled paths by descending luminance (from light to dark).

        This also changes the plot order in the way that all filled paths are plotted
        before polylines and outline paths.
        """
        fillings = []
        outlines = []
        current = Properties()
        props = self._properties
        for record in self._records:
            if record.type == RecordType.FILLED_PATHS:
                current = props.get(record.property_hash, current)
                key = colors.luminance(current.resolve_fill_color())
                fillings.append((key, record))
            else:
                outlines.append(record)

        fillings.sort(key=lambda r: r[0], reverse=True)
        records = [sort_rec[1] for sort_rec in fillings]
        records.extend(outlines)
        self._records = records


def placement_matrix(
    bbox: BoundingBox2d, sx: float = 1.0, sy: float = 1.0, rotation: float = 0.0
) -> Matrix44:
    """Returns a matrix to place the bbox in the first quadrant of the coordinate
    system (+x, +y).
    """
    if abs(sx) < 1e-9:
        sx = 1.0
    if abs(sy) < 1e-9:
        sy = 1.0
    m = Matrix44.scale(sx, sy, 1.0)
    if rotation:
        m @= Matrix44.z_rotate(math.radians(rotation))
    corners = m.fast_2d_transform(bbox.rect_vertices())
    tx, ty = BoundingBox2d(corners).extmin
    return m @ Matrix44.translate(-tx, -ty, 0)



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterable,
    Iterator,
    Union,
    cast,
    Sequence,
    Optional,
)
from itertools import chain
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
    merge_group_code_mappings,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER, VERTEXNAMES
from ezdxf.lldxf import const
from ezdxf.math import Vec3, Matrix44, NULLVEC, Z_AXIS, UVec
from ezdxf.math.transformtools import OCSTransform, NonUniformScalingError
from ezdxf.render.polyline import virtual_polyline_entities
from ezdxf.explode import explode_entity
from ezdxf.query import EntityQuery
from ezdxf.entities import factory
from ezdxf.audit import AuditError
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity, acdb_entity_group_codes
from .lwpolyline import FORMAT_CODES
from .subentity import LinkedEntities

if TYPE_CHECKING:
    from ezdxf.audit import Auditor
    from ezdxf.entities import DXFNamespace, Line, Arc, Face3d
    from ezdxf.layouts import BaseLayout
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.eztypes import FaceType

__all__ = ["Polyline", "Polyface", "Polymesh"]

acdb_polyline = DefSubclass(
    "AcDbPolylineDummy",
    {
        # AcDbPolylineDummy is a temporary solution while loading
        # Group code 66 is obsolete - Vertices follow flag
        # Elevation is a "dummy" point. The x and y values are always 0,
        # and the Z value is the polyline elevation:
        "elevation": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # Polyline flags (bit-coded):
        # 1 = closed POLYLINE or a POLYMESH closed in the M direction
        # 2 = Curve-fit vertices have been added
        # 4 = Spline-fit vertices have been added
        # 8 = 3D POLYLINE
        # 16 = POLYMESH
        # 32 = POLYMESH is closed in the N direction
        # 64 = POLYFACE
        # 128 = linetype pattern is generated continuously around the vertices
        "flags": DXFAttr(70, default=0),
        "default_start_width": DXFAttr(40, default=0, optional=True),
        "default_end_width": DXFAttr(41, default=0, optional=True),
        "m_count": DXFAttr(
            71,
            default=0,
            optional=True,
            validator=validator.is_greater_or_equal_zero,
            fixer=RETURN_DEFAULT,
        ),
        "n_count": DXFAttr(
            72,
            default=0,
            optional=True,
            validator=validator.is_greater_or_equal_zero,
            fixer=RETURN_DEFAULT,
        ),
        "m_smooth_density": DXFAttr(73, default=0, optional=True),
        "n_smooth_density": DXFAttr(74, default=0, optional=True),
        # Curves and smooth surface type:
        # 0 = No smooth surface fitted
        # 5 = Quadratic B-spline surface
        # 6 = Cubic B-spline surface
        # 8 = Bezier surface
        "smooth_type": DXFAttr(
            75,
            default=0,
            optional=True,
            validator=validator.is_one_of({0, 5, 6, 8}),
            fixer=RETURN_DEFAULT,
        ),
        "thickness": DXFAttr(39, default=0, optional=True),
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_polyline_group_codes = group_code_mapping(acdb_polyline, ignore=(66,))
merged_polyline_group_codes = merge_group_code_mappings(
    acdb_entity_group_codes, acdb_polyline_group_codes  # type: ignore
)


# Notes to SEQEND:
# todo: A loaded entity should have a valid SEQEND, a POLYLINE without vertices
#  makes no sense - has to be tested
#
# A virtual POLYLINE does not need a SEQEND, because it can not be exported,
# therefore the SEQEND entity should not be created in the
# DXFEntity.post_new_hook() method.
#
# A bounded POLYLINE needs a SEQEND to valid at export, therefore the
# LinkedEntities.post_bind_hook() method creates a new SEQEND after binding
# the entity to a document if needed.


@factory.register_entity
class Polyline(LinkedEntities):
    """DXF POLYLINE entity

    The POLYLINE entity is hard owner of its VERTEX entities and the SEQEND entity:

       VERTEX.dxf.owner == POLYLINE.dxf.handle
       SEQEND.dxf.owner == POLYLINE.dxf.handle

    """

    DXFTYPE = "POLYLINE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_polyline)
    # polyline flags (70)
    CLOSED = 1
    MESH_CLOSED_M_DIRECTION = CLOSED
    CURVE_FIT_VERTICES_ADDED = 2
    SPLINE_FIT_VERTICES_ADDED = 4
    POLYLINE_3D = 8
    POLYMESH = 16
    MESH_CLOSED_N_DIRECTION = 32
    POLYFACE = 64
    GENERATE_LINETYPE_PATTERN = 128
    # polymesh smooth type (75)
    NO_SMOOTH = 0
    QUADRATIC_BSPLINE = 5
    CUBIC_BSPLINE = 6
    BEZIER_SURFACE = 8
    ANY3D = POLYLINE_3D | POLYMESH | POLYFACE

    @property
    def vertices(self) -> list[DXFVertex]:
        return self._sub_entities  # type: ignore

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        """Loading interface. (internal API)"""
        # bypass DXFGraphic, loading proxy graphic is skipped!
        dxf = super(DXFGraphic, self).load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, merged_polyline_group_codes)
        return dxf

    def export_dxf(self, tagwriter: AbstractTagWriter):
        """Export POLYLINE entity and all linked entities: VERTEX, SEQEND."""
        super().export_dxf(tagwriter)
        # export sub-entities
        self.process_sub_entities(lambda e: e.export_dxf(tagwriter))

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export POLYLINE specific data as DXF tags."""
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, self.get_mode())

        tagwriter.write_tag2(66, 1)  # Vertices follow
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "elevation",
                "flags",
                "default_start_width",
                "default_end_width",
                "m_count",
                "n_count",
                "m_smooth_density",
                "n_smooth_density",
                "smooth_type",
                "thickness",
                "extrusion",
            ],
        )

    def on_layer_change(self, layer: str):
        """Event handler for layer change. Changes also the layer of all vertices.

        Args:
            layer: new layer as string

        """
        for v in self.vertices:
            v.dxf.layer = layer

    def on_linetype_change(self, linetype: str):
        """Event handler for linetype change. Changes also the linetype of all
        vertices.

        Args:
            linetype: new linetype as string

        """
        for v in self.vertices:
            v.dxf.linetype = linetype

    def get_vertex_flags(self) -> int:
        return const.VERTEX_FLAGS[self.get_mode()]

    def get_mode(self) -> str:
        """Returns POLYLINE type as string:

        - "AcDb2dPolyline"
        - "AcDb3dPolyline"
        - "AcDbPolygonMesh"
        - "AcDbPolyFaceMesh"

        """
        if self.is_3d_polyline:
            return "AcDb3dPolyline"
        elif self.is_polygon_mesh:
            return "AcDbPolygonMesh"
        elif self.is_poly_face_mesh:
            return "AcDbPolyFaceMesh"
        else:
            return "AcDb2dPolyline"

    @property
    def is_2d_polyline(self) -> bool:
        """``True`` if POLYLINE is a 2D polyline."""
        return self.dxf.flags & self.ANY3D == 0

    @property
    def is_3d_polyline(self) -> bool:
        """``True`` if POLYLINE is a 3D polyline."""
        return bool(self.dxf.flags & self.POLYLINE_3D)

    @property
    def is_polygon_mesh(self) -> bool:
        """``True`` if POLYLINE is a polygon mesh, see :class:`Polymesh`"""
        return bool(self.dxf.flags & self.POLYMESH)

    @property
    def is_poly_face_mesh(self) -> bool:
        """``True`` if POLYLINE is a poly face mesh, see :class:`Polyface`"""
        return bool(self.dxf.flags & self.POLYFACE)

    @property
    def is_closed(self) -> bool:
        """``True`` if POLYLINE is closed."""
        return bool(self.dxf.flags & self.CLOSED)

    @property
    def is_m_closed(self) -> bool:
        """``True`` if POLYLINE (as :class:`Polymesh`) is closed in m
        direction.
        """
        return bool(self.dxf.flags & self.MESH_CLOSED_M_DIRECTION)

    @property
    def is_n_closed(self) -> bool:
        """``True`` if POLYLINE (as :class:`Polymesh`) is closed in n
        direction.
        """
        return bool(self.dxf.flags & self.MESH_CLOSED_N_DIRECTION)

    @property
    def has_arc(self) -> bool:
        """Returns ``True`` if 2D POLYLINE has an arc segment."""
        if self.is_2d_polyline:
            return any(
                v.dxf.hasattr("bulge") and bool(v.dxf.bulge) for v in self.vertices
            )
        else:
            return False

    @property
    def has_width(self) -> bool:
        """Returns ``True`` if 2D POLYLINE has default width values or any
        segment with width attributes.
        """
        if self.is_2d_polyline:
            if self.dxf.hasattr("default_start_width") and bool(
                self.dxf.default_start_width
            ):
                return True
            if self.dxf.hasattr("default_end_width") and bool(
                self.dxf.default_end_width
            ):
                return True
            for v in self.vertices:
                if v.dxf.hasattr("start_width") and bool(v.dxf.start_width):
                    return True
                if v.dxf.hasattr("end_width") and bool(v.dxf.end_width):
                    return True
        return False

    def m_close(self, status=True) -> None:
        """Close POLYMESH in m direction if `status` is ``True`` (also closes
        POLYLINE), clears closed state if `status` is ``False``.
        """
        self.set_flag_state(self.MESH_CLOSED_M_DIRECTION, status, name="flags")

    def n_close(self, status=True) -> None:
        """Close POLYMESH in n direction if `status` is ``True``, clears closed
        state if `status` is ``False``.
        """
        self.set_flag_state(self.MESH_CLOSED_N_DIRECTION, status, name="flags")

    def close(self, m_close=True, n_close=False) -> None:
        """Set closed state of POLYMESH and POLYLINE in m direction and n
        direction. ``True`` set closed flag, ``False`` clears closed flag.
        """
        self.m_close(m_close)
        self.n_close(n_close)

    def __len__(self) -> int:
        """Returns count of :class:`Vertex` entities."""
        return len(self.vertices)

    def __getitem__(self, pos) -> DXFVertex:
        """Get :class:`Vertex` entity at position `pos`, supports list-like slicing."""
        return self.vertices[pos]

    def points(self) -> Iterator[Vec3]:
        """Returns all polyline points in :ref:`OCS` or :ref:`WCS` coordinates as
        :class:`~ezdxf.math.Vec3`.

        These are the raw location coordinates stored in the :class:`Vertex` entities.
        A separately stored elevation value will not be applied. The points of
        2D polylines are :ref:`OCS` coordinates other polyline types return :ref:`WCS`
        coordinates.
        """
        return (vertex.dxf.location for vertex in self.vertices)

    def points_in_wcs(self) -> Iterator[Vec3]:
        """Returns all polyline points in :ref:`WCS` coordinates as
        :class:`~ezdxf.math.Vec3`.

        .. versionadded:: 1.4

        """
        points = self.points()
        if not self.is_2d_polyline:
            return points

        elevation: float = self.dxf.elevation.z
        if elevation:
            points = (p.replace(z=elevation) for p in points)

        if Z_AXIS.isclose(self.dxf.extrusion):  # OCS == WCS
            return points

        ocs = self.ocs()
        return ocs.points_to_wcs(points)

    def _append_vertex(self, vertex: DXFVertex) -> None:
        self.vertices.append(vertex)

    def append_vertices(self, points: Iterable[UVec], dxfattribs=None) -> None:
        """Append multiple :class:`Vertex` entities at location `points`.

        Args:
            points: iterable of (x, y[, z]) tuples
            dxfattribs: dict of DXF attributes for the VERTEX objects

        """
        dxfattribs = dict(dxfattribs or {})
        for vertex in self._build_dxf_vertices(points, dxfattribs):
            self._append_vertex(vertex)

    def append_formatted_vertices(
        self,
        points: Iterable[Sequence],
        format: str = "xy",
        dxfattribs=None,
    ) -> None:
        """Append multiple :class:`Vertex` entities at location `points`.

        Args:
            points: iterable of (x, y, [start_width, [end_width, [bulge]]])
                    tuple
            format: format string, default is "xy", see: :ref:`format codes`
            dxfattribs: dict of DXF attributes for the VERTEX objects

        """
        dxfattribs = dict(dxfattribs or {})
        dxfattribs["flags"] = dxfattribs.get("flags", 0) | self.get_vertex_flags()

        # same DXF attributes for VERTEX entities as for POLYLINE
        dxfattribs["owner"] = self.dxf.owner
        dxfattribs["layer"] = self.dxf.layer
        if self.dxf.hasattr("linetype"):
            dxfattribs["linetype"] = self.dxf.linetype

        for point in points:
            attribs = vertex_attribs(point, format)
            attribs.update(dxfattribs)
            vertex = cast(DXFVertex, self._new_compound_entity("VERTEX", attribs))
            self._append_vertex(vertex)

    def append_vertex(self, point: UVec, dxfattribs=None) -> None:
        """Append a single :class:`Vertex` entity at location `point`.

        Args:
            point: as (x, y[, z]) tuple
            dxfattribs: dict of DXF attributes for :class:`Vertex` class

        """
        dxfattribs = dict(dxfattribs or {})
        for vertex in self._build_dxf_vertices([point], dxfattribs):
            self._append_vertex(vertex)

    def insert_vertices(
        self, pos: int, points: Iterable[UVec], dxfattribs=None
    ) -> None:
        """Insert vertices `points` into :attr:`Polyline.vertices` list
        at insertion location `pos` .

        Args:
            pos: insertion position of list :attr:`Polyline.vertices`
            points: list of (x, y[, z]) tuples
            dxfattribs: dict of DXF attributes for :class:`Vertex` class

        """
        dxfattribs = dict(dxfattribs or {})
        self.vertices[pos:pos] = list(self._build_dxf_vertices(points, dxfattribs))

    def _build_dxf_vertices(
        self, points: Iterable[UVec], dxfattribs: dict
    ) -> Iterator[DXFVertex]:
        """Converts point (x, y, z) tuples into DXFVertex objects.

        Args:
            points: list of (x, y, z)-tuples
            dxfattribs: dict of DXF attributes
        """
        dxfattribs["flags"] = dxfattribs.get("flags", 0) | self.get_vertex_flags()

        # same DXF attributes for VERTEX entities as for POLYLINE
        dxfattribs["owner"] = self.dxf.handle
        dxfattribs["layer"] = self.dxf.layer
        if self.dxf.hasattr("linetype"):
            dxfattribs["linetype"] = self.dxf.linetype
        for point in points:
            dxfattribs["location"] = Vec3(point)
            yield cast(DXFVertex, self._new_compound_entity("VERTEX", dxfattribs))

    def cast(self) -> Union[Polyline, Polymesh, Polyface]:
        mode = self.get_mode()
        if mode == "AcDbPolyFaceMesh":
            return Polyface.from_polyline(self)
        elif mode == "AcDbPolygonMesh":
            return Polymesh.from_polyline(self)
        else:
            return self

    def transform(self, m: Matrix44) -> Polyline:
        """Transform the POLYLINE entity by transformation matrix `m` inplace.

        A non-uniform scaling is not supported if a 2D POLYLINE contains
        circular arc segments (bulges).

        Args:
            m: transformation :class:`~ezdxf.math.Matrix44`

        Raises:
            NonUniformScalingError: for non-uniform scaling of 2D POLYLINE
                containing circular arc segments (bulges)

        """

        def _ocs_locations(elevation):
            for vertex in self.vertices:
                location = vertex.dxf.location
                if elevation is not None:
                    # Older DXF versions may not have written the z-axis, so
                    # replace existing z-axis by the elevation value.
                    location = location.replace(z=elevation)
                yield location

        if self.is_2d_polyline:
            dxf = self.dxf
            ocs = OCSTransform(self.dxf.extrusion, m)
            if not ocs.scale_uniform and self.has_arc:
                raise NonUniformScalingError(
                    "2D POLYLINE containing arcs (bulges) does not support non uniform scaling"
                )
                # The caller function has to catch this exception and explode the
                # 2D POLYLINE into LINE and ELLIPSE entities.
            if dxf.hasattr("elevation"):
                z_axis = dxf.elevation.z
            else:
                z_axis = None
            vertices = [
                ocs.transform_vertex(vertex) for vertex in _ocs_locations(z_axis)
            ]

            # All vertices of a 2D polyline must have the same z-axis, which is
            # the elevation of the polyline:
            if vertices:
                dxf.elevation = vertices[0].replace(x=0.0, y=0.0)

            for vertex, location in zip(self.vertices, vertices):
                vdxf = vertex.dxf
                vdxf.location = location
                if vdxf.hasattr("start_width"):
                    vdxf.start_width = ocs.transform_width(vdxf.start_width)
                if vdxf.hasattr("end_width"):
                    vdxf.end_width = ocs.transform_width(vdxf.end_width)

            if dxf.hasattr("default_start_width"):
                dxf.default_start_width = ocs.transform_width(dxf.default_start_width)
            if dxf.hasattr("default_end_width"):
                dxf.default_end_width = ocs.transform_width(dxf.default_end_width)
            if dxf.hasattr("thickness"):
                dxf.thickness = ocs.transform_thickness(dxf.thickness)

            dxf.extrusion = ocs.new_extrusion
        else:
            for vertex in self.vertices:
                vertex.transform(m)
        self.post_transform(m)
        return self

    def explode(self, target_layout: Optional[BaseLayout] = None) -> EntityQuery:
        """Explode the POLYLINE entity as DXF primitives (LINE, ARC or 3DFACE)
        into the target layout, if the target layout is ``None``, the target
        layout is the layout of the POLYLINE entity.

        Returns an :class:`~ezdxf.query.EntityQuery` container referencing all
        DXF primitives.

        Args:
            target_layout: target layout for DXF primitives, ``None`` for same
                layout as source entity.

        """
        return explode_entity(self, target_layout)

    def virtual_entities(self) -> Iterator[Union[Line, Arc, Face3d]]:
        """Yields the graphical representation of POLYLINE as virtual DXF
        primitives (LINE, ARC or 3DFACE).

        These virtual entities are located at the original location, but are not
        stored in the entity database, have no handle and are not assigned to
        any layout.

        """
        for e in virtual_polyline_entities(self):
            e.set_source_of_copy(self)
            yield e

    def audit(self, auditor: Auditor) -> None:
        """Audit and repair the POLYLINE entity."""

        def audit_sub_entity(entity):
            entity.doc = doc  # grant same document
            dxf = entity.dxf
            if dxf.owner != owner:
                dxf.owner = owner
            if dxf.layer != layer:
                dxf.layer = layer

        doc = self.doc
        owner = self.dxf.handle
        layer = self.dxf.layer
        for vertex in self.vertices:
            audit_sub_entity(vertex)

        seqend = self.seqend
        if seqend:
            audit_sub_entity(seqend)
        elif doc:
            self.new_seqend()
            auditor.fixed_error(
                code=AuditError.MISSING_REQUIRED_SEQEND,
                message=f"Created required SEQEND entity for {str(self)}.",
                dxf_entity=self,
            )


class Polyface(Polyline):
    """
    PolyFace structure:

    POLYLINE
      AcDbEntity
      AcDbPolyFaceMesh
    VERTEX - Vertex
      AcDbEntity
      AcDbVertex
      AcDbPolyFaceMeshVertex
    VERTEX - Face
      AcDbEntity
      AcDbFaceRecord
    SEQEND

    Order of mesh_vertices and face_records is important (DXF R2010):

        1. mesh_vertices: the polyface mesh vertex locations
        2. face_records: indices of the face forming vertices

    """

    @classmethod
    def from_polyline(cls, polyline: Polyline) -> Polyface:
        polyface = cls.shallow_copy(polyline)
        polyface._sub_entities = polyline._sub_entities
        polyface.seqend = polyline.seqend
        # do not destroy polyline - all data would be lost
        return polyface

    def append_face(self, face: FaceType, dxfattribs=None) -> None:
        """Append a single face. A `face` is a sequence of (x, y, z) tuples.

        Args:
            face: sequence of (x, y, z) tuples
            dxfattribs: dict of DXF attributes for the VERTEX objects

        """
        self.append_faces([face], dxfattribs)

    def _points_to_dxf_vertices(
        self, points: Iterable[UVec], dxfattribs
    ) -> list[DXFVertex]:
        """Convert (x, y, z) tuples into DXFVertex objects.

        Args:
            points: sequence of (x, y, z) tuples
            dxfattribs: dict of DXF attributes for the VERTEX entity

        """
        dxfattribs["flags"] = dxfattribs.get("flags", 0) | self.get_vertex_flags()

        # All vertices have to be on the same layer as the POLYLINE entity:
        dxfattribs["layer"] = self.get_dxf_attrib("layer", "0")
        vertices: list[DXFVertex] = []
        for point in points:
            dxfattribs["location"] = point
            vertices.append(
                cast("DXFVertex", self._new_compound_entity("VERTEX", dxfattribs))
            )
        return vertices

    def append_faces(self, faces: Iterable[FaceType], dxfattribs=None) -> None:
        """Append multiple `faces`. `faces` is a list of single faces and a
        single face is a sequence of (x, y, z) tuples.

        Args:
            faces: iterable of sequences of (x, y, z) tuples
            dxfattribs: dict of DXF attributes for the VERTEX entity

        """

        def new_face_record():
            dxfattribs["flags"] = const.VTX_3D_POLYFACE_MESH_VERTEX
            # location of face record vertex is always (0, 0, 0)
            dxfattribs["location"] = Vec3()
            return cast(DXFVertex, self._new_compound_entity("VERTEX", dxfattribs))

        dxfattribs = dict(dxfattribs or {})

        existing_vertices, existing_faces = self.indexed_faces()
        new_faces: list[FaceProxy] = []
        for face in faces:
            face_mesh_vertices = self._points_to_dxf_vertices(face, {})
            # Index of first new vertex
            index = len(existing_vertices)
            existing_vertices.extend(face_mesh_vertices)
            face_record = FaceProxy(new_face_record(), existing_vertices)

            # Set VERTEX indices:
            face_record.indices = tuple(range(index, index + len(face_mesh_vertices)))
            new_faces.append(face_record)
        self._rebuild(chain(existing_faces, new_faces))

    def _rebuild(self, faces: Iterable[FaceProxy], precision: int = 6) -> None:
        """Build a valid POLYFACE structure from `faces`.

        Args:
            faces: iterable of FaceProxy objects.

        """
        polyface_builder = PolyfaceBuilder(faces, precision=precision)
        # Why is list[DXFGraphic] incompatible to list[DXFVertex] when DXFVertex
        # is a subclass of DXFGraphic?
        self._sub_entities = polyface_builder.get_vertices()  # type: ignore
        self.update_count(polyface_builder.nvertices, polyface_builder.nfaces)

    def update_count(self, nvertices: int, nfaces: int) -> None:
        self.dxf.m_count = nvertices
        self.dxf.n_count = nfaces

    def optimize(self, precision: int = 6) -> None:
        """Rebuilds the :class:`Polyface` by merging vertices with nearly same vertex
        locations.

        Args:
            precision: floating point precision for determining identical
                vertex locations

        """
        vertices, faces = self.indexed_faces()
        self._rebuild(faces, precision)

    def faces(self) -> Iterator[list[DXFVertex]]:
        """Iterable of all faces, a face is a tuple of vertices.

        Returns:
             list of [vertex, vertex, vertex, [vertex,] face_record]

        """
        _, faces = self.indexed_faces()
        for face in faces:
            face_vertices = list(face)
            face_vertices.append(face.face_record)
            yield face_vertices

    def indexed_faces(self) -> tuple[list[DXFVertex], Iterator[FaceProxy]]:
        """Returns a list of all vertices and a generator of FaceProxy()
        objects.

        (internal API)
        """
        vertices: list[DXFVertex] = []
        face_records: list[DXFVertex] = []
        for vertex in self.vertices:
            (vertices if vertex.is_poly_face_mesh_vertex else face_records).append(
                vertex
            )

        faces = (FaceProxy(face_record, vertices) for face_record in face_records)
        return vertices, faces


class FaceProxy:
    """Represents a single face of a polyface structure. (internal class)

    vertices:

        List of all polyface vertices.

    face_record:

        The face forming vertex of type ``AcDbFaceRecord``, contains the indices
        to the face building vertices. Indices of the DXF structure are 1-based
        and a negative index indicates the beginning of an invisible edge.
        Face.face_record.dxf.color determines the color of the face.

    indices:

        Indices to the face building vertices as tuple. This indices are 0-base
        and are used to get vertices from the list `Face.vertices`.

    """

    __slots__ = ("vertices", "face_record", "indices")

    def __init__(self, face_record: DXFVertex, vertices: Sequence[DXFVertex]):
        """Returns iterable of all face vertices as :class:`Vertex` entities."""
        self.vertices: Sequence[DXFVertex] = vertices
        self.face_record: DXFVertex = face_record
        self.indices: Sequence[int] = self._indices()

    def __len__(self) -> int:
        """Returns count of face vertices (without face_record)."""
        return len(self.indices)

    def __getitem__(self, pos: int) -> DXFVertex:
        """Returns :class:`Vertex` at position `pos`.

        Args:
            pos: vertex position 0-based

        """
        return self.vertices[self.indices[pos]]

    def __iter__(self) -> Iterator["DXFVertex"]:
        return (self.vertices[index] for index in self.indices)

    def points(self) -> Iterator[UVec]:
        """Returns iterable of all face vertex locations as (x, y, z)-tuples."""
        return (vertex.dxf.location for vertex in self)

    def _raw_indices(self) -> Iterable[int]:
        return (self.face_record.get_dxf_attrib(name, 0) for name in const.VERTEXNAMES)

    def _indices(self) -> Sequence[int]:
        return tuple(abs(index) - 1 for index in self._raw_indices() if index != 0)

    def is_edge_visible(self, pos: int) -> bool:
        """Returns ``True`` if edge starting at vertex `pos` is visible.

        Args:
            pos: vertex position 0-based

        """
        name = const.VERTEXNAMES[pos]
        return self.face_record.get_dxf_attrib(name) > 0


class PolyfaceBuilder:
    """Optimized POLYFACE builder. (internal class)"""

    def __init__(self, faces: Iterable[FaceProxy], precision: int = 6):
        self.precision: int = precision
        self.faces: list[DXFVertex] = []
        self.vertices: list[DXFVertex] = []
        self.index_mapping: dict[tuple[float, ...], int] = {}
        self.build(faces)

    @property
    def nvertices(self) -> int:
        return len(self.vertices)

    @property
    def nfaces(self) -> int:
        return len(self.faces)

    def get_vertices(self) -> list[DXFVertex]:
        vertices = self.vertices[:]
        vertices.extend(self.faces)
        return vertices

    def build(self, faces: Iterable[FaceProxy]) -> None:
        for face in faces:
            face_record = face.face_record
            for vertex, name in zip(face, VERTEXNAMES):
                index = self.add(vertex)
                # preserve sign of old index value
                sign = -1 if face_record.dxf.get(name, 0) < 0 else +1
                face_record.dxf.set(name, (index + 1) * sign)
            self.faces.append(face_record)

    def add(self, vertex: DXFVertex) -> int:
        def key(point):
            return tuple((round(coord, self.precision) for coord in point))

        location = key(vertex.dxf.location)
        try:
            return self.index_mapping[location]
        except KeyError:
            index = len(self.vertices)
            self.index_mapping[location] = index
            self.vertices.append(vertex)
            return index


class Polymesh(Polyline):
    """
    PolyMesh structure:

    POLYLINE
      AcDbEntity
      AcDbPolygonMesh
    VERTEX
      AcDbEntity
      AcDbVertex
      AcDbPolygonMeshVertex
    """

    @classmethod
    def from_polyline(cls, polyline: Polyline) -> Polymesh:
        polymesh = cls.shallow_copy(polyline)
        polymesh._sub_entities = polyline._sub_entities
        polymesh.seqend = polyline.seqend
        return polymesh

    def set_mesh_vertex(self, pos: tuple[int, int], point: UVec, dxfattribs=None):
        """Set location and DXF attributes of a single mesh vertex.

        Args:
            pos: 0-based (row, col) tuple, position of mesh vertex
            point: (x, y, z) tuple, new 3D coordinates of the mesh vertex
            dxfattribs: dict of DXF attributes

        """
        dxfattribs = dict(dxfattribs or {})
        dxfattribs["location"] = point
        vertex = self.get_mesh_vertex(pos)
        vertex.update_dxf_attribs(dxfattribs)

    def get_mesh_vertex(self, pos: tuple[int, int]) -> DXFVertex:
        """Get location of a single mesh vertex.

        Args:
            pos: 0-based (row, col) tuple, position of mesh vertex

        """
        m_count = self.dxf.m_count
        n_count = self.dxf.n_count
        m, n = pos
        if 0 <= m < m_count and 0 <= n < n_count:
            return self.vertices[m * n_count + n]
        else:
            raise const.DXFIndexError(repr(pos))

    def get_mesh_vertex_cache(self) -> MeshVertexCache:
        """Get a :class:`MeshVertexCache` object for this POLYMESH.
        The caching object provides fast access to the :attr:`location`
        attribute of mesh vertices.

        """
        return MeshVertexCache(self)


class MeshVertexCache:
    """Cache mesh vertices in a dict, keys are 0-based (row, col)-tuples.

    vertices:
        Dict of mesh vertices, keys are 0-based (row, col)-tuples. Writing to
        this dict doesn't change the DXF entity.

    """

    __slots__ = ("vertices",)

    def __init__(self, mesh: Polyline):
        self.vertices: dict[tuple[int, int], DXFVertex] = self._setup(
            mesh, mesh.dxf.m_count, mesh.dxf.n_count
        )

    def _setup(self, mesh: Polyline, m_count: int, n_count: int) -> dict:
        cache: dict[tuple[int, int], DXFVertex] = {}
        vertices = iter(mesh.vertices)
        for m in range(m_count):
            for n in range(n_count):
                cache[(m, n)] = next(vertices)
        return cache

    def __getitem__(self, pos: tuple[int, int]) -> UVec:
        """Get mesh vertex location as (x, y, z)-tuple.

        Args:
            pos: 0-based (row, col)-tuple.

        """
        try:
            return self.vertices[pos].dxf.location
        except KeyError:
            raise const.DXFIndexError(repr(pos))

    def __setitem__(self, pos: tuple[int, int], location: UVec) -> None:
        """Get mesh vertex location as (x, y, z)-tuple.

        Args:
            pos: 0-based (row, col)-tuple.
            location: (x, y, z)-tuple

        """
        try:
            self.vertices[pos].dxf.location = location
        except KeyError:
            raise const.DXFIndexError(repr(pos))


acdb_vertex = DefSubclass(
    "AcDbVertex",
    {  # last subclass index -1
        # Location point in OCS if 2D, and WCS if 3D
        "location": DXFAttr(10, xtype=XType.point3d),
        "start_width": DXFAttr(40, default=0, optional=True),
        "end_width": DXFAttr(41, default=0, optional=True),
        # Bulge (optional; default is 0). The bulge is the tangent of one fourth
        # the included angle for an arc segment, made negative if the arc goes
        # clockwise from the start point to the endpoint. A bulge of 0 indicates
        # a straight segment, and a bulge of 1 is a semicircle.
        "bulge": DXFAttr(42, default=0, optional=True),
        "flags": DXFAttr(70, default=0),
        # Curve fit tangent direction (in degrees)
        "tangent": DXFAttr(50, optional=True),
        "vtx0": DXFAttr(71, optional=True),
        "vtx1": DXFAttr(72, optional=True),
        "vtx2": DXFAttr(73, optional=True),
        "vtx3": DXFAttr(74, optional=True),
        "vertex_identifier": DXFAttr(91, optional=True),
    },
)
acdb_vertex_group_codes = group_code_mapping(acdb_vertex)
merged_vertex_group_codes = merge_group_code_mappings(
    acdb_entity_group_codes, acdb_vertex_group_codes  # type: ignore
)


@factory.register_entity
class DXFVertex(DXFGraphic):
    """DXF VERTEX entity"""

    DXFTYPE = "VERTEX"

    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_vertex)
    # Extra vertex created by curve-fitting:
    EXTRA_VERTEX_CREATED = 1

    # Curve-fit tangent defined for this vertex. A curve-fit tangent direction
    # of 0 may be omitted from the DXF output, but is significant if this bit
    # is set:
    CURVE_FIT_TANGENT = 2

    # 4 = unused, never set in dxf files
    # Spline vertex created by spline-fitting
    SPLINE_VERTEX_CREATED = 8
    SPLINE_FRAME_CONTROL_POINT = 16
    POLYLINE_3D_VERTEX = 32
    POLYGON_MESH_VERTEX = 64
    POLYFACE_MESH_VERTEX = 128
    FACE_FLAGS = POLYGON_MESH_VERTEX + POLYFACE_MESH_VERTEX
    VTX3D = POLYLINE_3D_VERTEX + POLYGON_MESH_VERTEX + POLYFACE_MESH_VERTEX

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        """Loading interface. (internal API)"""
        # bypass DXFGraphic, loading proxy graphic is skipped!
        dxf = super(DXFGraphic, self).load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, merged_vertex_group_codes)
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        # VERTEX can have 3 subclasses if representing a `face record` or
        # 4 subclasses if representing a vertex location, just the last
        # subclass contains data
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            if self.is_face_record:
                tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbFaceRecord")
            else:
                tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbVertex")
                if self.is_3d_polyline_vertex:
                    tagwriter.write_tag2(SUBCLASS_MARKER, "AcDb3dPolylineVertex")
                elif self.is_poly_face_mesh_vertex:
                    tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbPolyFaceMeshVertex")
                elif self.is_polygon_mesh_vertex:
                    tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbPolygonMeshVertex")
                else:
                    tagwriter.write_tag2(SUBCLASS_MARKER, "AcDb2dVertex")

        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "location",
                "start_width",
                "end_width",
                "bulge",
                "flags",
                "tangent",
                "vtx0",
                "vtx1",
                "vtx2",
                "vtx3",
                "vertex_identifier",
            ],
        )

    @property
    def is_2d_polyline_vertex(self) -> bool:
        return self.dxf.flags & self.VTX3D == 0

    @property
    def is_3d_polyline_vertex(self) -> bool:
        return self.dxf.flags & self.POLYLINE_3D_VERTEX

    @property
    def is_polygon_mesh_vertex(self) -> bool:
        return self.dxf.flags & self.POLYGON_MESH_VERTEX

    @property
    def is_poly_face_mesh_vertex(self) -> bool:
        return self.dxf.flags & self.FACE_FLAGS == self.FACE_FLAGS

    @property
    def is_face_record(self) -> bool:
        return (self.dxf.flags & self.FACE_FLAGS) == self.POLYFACE_MESH_VERTEX

    def transform(self, m: Matrix44) -> DXFVertex:
        """Transform the VERTEX entity by transformation matrix `m` inplace."""
        if self.is_face_record:
            return self
        self.dxf.location = m.transform(self.dxf.location)
        return self

    def format(self, format="xyz") -> Sequence:
        """Return formatted vertex components as tuple.

        Format codes:

            - ``x`` = x-coordinate
            - ``y`` = y-coordinate
            - ``z`` = z-coordinate
            - ``s`` = start width
            - ``e`` = end width
            - ``b`` = bulge value
            - ``v`` = (x, y, z) as tuple

        Args:
            format: format string, default is "xyz"

        """
        dxf = self.dxf
        v = Vec3(dxf.location)
        x, y, z = v.xyz
        b = dxf.bulge
        s = dxf.start_width
        e = dxf.end_width
        vars = locals()
        return tuple(vars[code] for code in format.lower())


def vertex_attribs(data: Sequence, format="xyseb") -> dict:
    """Create VERTEX attributes from input data.

    Format codes:

        - ``x`` = x-coordinate
        - ``y`` = y-coordinate
        - ``s`` = start width
        - ``e`` = end width
        - ``b`` = bulge value
        - ``v`` = (x, y [,z]) tuple (z-axis is ignored)

    Args:
        data: list or tuple of point components
        format: format string, default is 'xyseb'

    Returns:
       dict with keys: 'location', 'bulge', 'start_width', 'end_width'

    """
    attribs = dict()
    format = [code for code in format.lower() if code in FORMAT_CODES]
    location = Vec3()
    for code, value in zip(format, data):
        if code not in FORMAT_CODES:
            continue
        if code == "v":
            location = Vec3(value)
        elif code == "b":
            attribs["bulge"] = float(value)
        elif code == "s":
            attribs["start_width"] = float(value)
        elif code == "e":
            attribs["end_width"] = float(value)
        elif code == "x":
            location = location.replace(x=float(value))
        elif code == "y":
            location = location.replace(y=float(value))
    attribs["location"] = location
    return attribs



#  Copyright (c) 2021-2023, Manfred Moitzi
#  License: MIT License
import sys
import argparse
from pathlib import Path
from ezdxf import options, print_config
from ezdxf import commands
from ezdxf.fonts import fonts

YES_NO = {True: "yes", False: "no"}
options.set(options.CORE, "LOAD_PROXY_GRAPHICS", "true")


def add_common_arguments(parser):
    parser.add_argument(
        "-V",
        "--version",
        action="store_true",
        help="show version and exit",
    )
    parser.add_argument(
        "-f",
        "--fonts",
        action="store_true",
        help="rebuild system font cache and print all fonts found",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="give more output",
    )
    parser.add_argument(
        "--config",
        action="store",
        help="path to a config file",
    )
    parser.add_argument(
        "--log",
        action="store",
        help='path to a verbose appending log, "stderr" logs to the '
        "standard error stream",
    )


def print_version(verbose=False):
    print_config(verbose=verbose, stream=sys.stdout)


def print_available_fonts(verbose=False):
    from ezdxf.fonts import fonts
    print("Rebuilding system font cache.")
    fonts.build_system_font_cache()
    fonts.font_manager.print_available_fonts(verbose)


def setup_log(args):
    import logging
    from datetime import datetime
    from io import StringIO

    level = "DEBUG" if args.verbose else "INFO"
    if args.log.lower() == "stderr":
        logging.basicConfig(stream=sys.stderr, level=level)
    else:
        logging.basicConfig(filename=args.log, level=level)
    print(f'Appending logs to file "{args.log}", logging level: {level}\n')
    logger = logging.getLogger("ezdxf")
    logger.info("***** Launch time: " + datetime.now().isoformat() + " *****")
    if args.verbose:
        s = StringIO()
        print_config(verbose=True, stream=s)
        logger.info("configuration\n" + s.getvalue())


DESCRIPTION = """
Command launcher for the Python package "ezdxf": https://pypi.org/project/ezdxf/

"""


def main():
    parser = argparse.ArgumentParser(
        "ezdxf",
        description=DESCRIPTION,
    )
    add_common_arguments(parser)
    subparsers = parser.add_subparsers(dest="command")
    commands.add_parsers(subparsers)

    args = parser.parse_args(sys.argv[1:])
    help_ = True
    if args.config:
        config = Path(args.config)
        if config.exists():
            options.read_file(args.config)
            if args.verbose:
                print(f'using config file: "{config}"')
        else:
            print(f'config file "{config}" not found')
    if args.log:
        setup_log(args)
    if args.version:
        print_version(verbose=args.verbose)
        help_ = False
    if args.fonts:
        print_available_fonts(args.verbose)
        help_ = False

    run = commands.get(args.command)
    if run:
        # For the case automatic font loading is disabled:
        fonts.load()
        run(args)
    elif help_:
        parser.print_help()


if __name__ == "__main__":
    main()



#  Copyright (c) 2022-2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import Iterator, Sequence

from ezdxf.acis.sat import parse_sat, SatEntity
from ezdxf.acis.sab import parse_sab, SabEntity


class AcisData:
    def __init__(self, name: str = "unknown", handle: str = ""):
        self.lines: list[str] = []
        self.name: str = name
        self.handle: str = handle


class BinaryAcisData(AcisData):
    def __init__(self, data: bytes, name: str, handle: str):
        super().__init__(name, handle)
        self.lines = list(make_sab_records(data))


class TextAcisData(AcisData):
    def __init__(self, data: Sequence[str], name: str, handle: str):
        super().__init__(name, handle)
        self.lines = list(make_sat_records(data))


def ptr_str(e):
    return "~" if e.is_null_ptr else str(e)


def make_sat_records(data: Sequence[str]) -> Iterator[str]:
    builder = parse_sat(data)
    yield from builder.header.dumps()
    builder.reset_ids()
    for entity in builder.entities:
        content = [str(entity)]
        content.append(ptr_str(entity.attributes))
        for field in entity.data:
            if isinstance(field, SatEntity):
                content.append(ptr_str(field))
            else:
                content.append(field)
        yield " ".join(content)


def make_sab_records(data: bytes) -> Iterator[str]:
    builder = parse_sab(data)
    yield from builder.header.dumps()
    builder.reset_ids()
    for entity in builder.entities:
        content = [str(entity)]
        content.append(ptr_str(entity.attributes))
        for tag in entity.data:
            if isinstance(tag.value, SabEntity):
                content.append(ptr_str(tag.value))
            else:
                content.append(f"{tag.value}<{tag.tag}>")
        yield " ".join(content)



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.acad_xrec_roundtrip import RoundtripXRecord, find_section
from ezdxf.entities import XRecord
from ezdxf.lldxf.tags import Tags


@pytest.fixture
def xrecord():
    xrec = XRecord()
    xrec.tags = Tags.from_text(SECTIONS)
    return xrec


def test_init():
    xrec = RoundtripXRecord()
    assert xrec.xrecord is not None


def test_has_section(xrecord):
    rtr = RoundtripXRecord(xrecord)
    assert rtr.has_section("ACAD_XXX") is True
    assert rtr.has_section("ACAD_AAA") is False


def test_add_new_section():
    rtr = RoundtripXRecord()
    rtr.set_section("ACAD_XXX", Tags.from_tuples([(1, "TEXT")]))
    assert rtr.xrecord.tags == [(102, "ACAD_XXX"), (1, "TEXT")]


def test_replace_existing_section(xrecord):
    rtr = RoundtripXRecord(xrecord)
    rtr.set_section("ACAD_XXX", Tags.from_tuples([(70, 7)]))
    assert xrecord.tags == [(102, "ACAD_XXX"), (70, 7), (102, "ACAD_YYY"), (40, 2.0)]


def test_get_section(xrecord):
    rtr = RoundtripXRecord(xrecord)
    assert rtr.get_section("ACAD_XXX") == [(40, 1.0), (1, "TEXT")]
    assert rtr.get_section("ACAD_YYY") == [(40, 2.0)]


def test_discard_existing_section(xrecord):
    rtr = RoundtripXRecord(xrecord)
    rtr.discard("ACAD_XXX")
    assert xrecord.tags == [(102, "ACAD_YYY"), (40, 2.0)]


def test_discard_non_existing_section(xrecord):
    rtr = RoundtripXRecord(xrecord)
    rtr.discard("ACAD_AAA")
    assert xrecord.tags == [
        (102, "ACAD_XXX"),
        (40, 1.0),
        (1, "TEXT"),
        (102, "ACAD_YYY"),
        (40, 2.0),
    ]


def test_locate_section():
    tags = Tags.from_text(SECTIONS)
    start, end = find_section(tags, "ACAD")
    assert start == -1 and end == -1
    start, end = find_section(tags, "ACAD_XXX")
    assert start == 0 and end == 3
    start, end = find_section(tags, "ACAD_YYY")
    assert start == 3 and end == 5


SECTIONS = """102
ACAD_XXX
40
1.0
1
TEXT
102
ACAD_YYY
40
2.0
"""

if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest
import pickle
from ezdxf.math import Vec3, Vec2, Matrix44, close_vectors

# Import from 'ezdxf.math._bezier3p' to test Python implementation
from ezdxf.math._bezier3p import Bezier3P
from ezdxf.acc import USE_C_EXT

curve_classes = [Bezier3P]
if USE_C_EXT:
    from ezdxf.acc.bezier3p import Bezier3P as CBezier3P

    curve_classes.append(CBezier3P)

DEFPOINTS2D = Vec2.list([(0.0, 0.0), (5.0, 5.0), (10.0, 0.0)])
DEFPOINTS3D = Vec3.list([(0.0, 0.0, 0.0), (50.0, 50.0, 50.0), (100.0, 0.0, 0.0)])


@pytest.fixture(params=curve_classes)
def bezier(request):
    return request.param


def test_accepts_2d_points(bezier):
    curve = bezier(DEFPOINTS2D)
    for index, chk in enumerate(Vec2.generate(POINTS2D)):
        assert curve.point(index * 0.1).isclose(chk)


def test_objects_are_immutable(bezier):
    curve = bezier(DEFPOINTS3D)
    with pytest.raises(TypeError):
        curve.control_points[0] = (1, 2, 3)


def test_approximate(bezier):
    curve = bezier(DEFPOINTS2D)
    with pytest.raises(ValueError):
        list(curve.approximate(0))
    assert list(curve.approximate(1)) == [DEFPOINTS2D[0], DEFPOINTS2D[-1]]
    assert list(curve.approximate(2)) == [
        POINTS2D[0],
        POINTS2D[5],
        POINTS2D[-1],
    ]


def test_first_derivative(bezier):
    dbcurve = bezier(DEFPOINTS2D)
    for index, chk in enumerate(Vec2.generate(TANGENTS2D)):
        assert dbcurve.tangent(index * 0.1).isclose(chk)


def test_reverse_points(bezier):
    curve = bezier(DEFPOINTS2D)
    vertices = list(curve.approximate(10))
    rev_curve = curve.reverse()
    rev_vertices = list(rev_curve.approximate(10))
    assert close_vectors(reversed(vertices), rev_vertices)


def test_transformation_interface(bezier):
    curve = bezier(DEFPOINTS3D)
    new = curve.transform(Matrix44.translate(1, 2, 3))
    assert new.control_points[0] == Vec3(DEFPOINTS3D[0]) + (1, 2, 3)
    assert (
        new.control_points[0] != curve.control_points[0]
    ), "expected a new object"


def test_transformation_returns_always_3d_curves(bezier):
    curve = bezier(DEFPOINTS2D)
    new = curve.transform(Matrix44.translate(1, 2, 3))
    assert len(new.control_points[0]) == 3


def test_flattening(bezier):
    curve = bezier(DEFPOINTS2D)
    assert len(list(curve.flattening(1.0, segments=4))) == 5
    assert len(list(curve.flattening(0.1, segments=4))) == 9


@pytest.mark.parametrize(
    "z",
    [
        1e99,
        1e79,
        1e59,
        1e39,
        1e19,
        1e9,
        1e6,
        1000,
        0,
        -1e99,
        -1e79,
        -1e59,
        -1e39,
        -1e19,
        -1e9,
        -1e6,
        -1000,
    ],
)
def test_flattening_big_z_coordinates(bezier, z):
    """Test based on issue #574"""
    cp = Vec3.list([
        (888, 770, z),
        (887, 623, z),
        (901, 478, z),
    ])
    curve = bezier(cp)
    points = list(curve.flattening(0.01))
    # Don't care about the result, it should just not break!
    assert len(points) > 0


def test_approximated_length(bezier):
    length = bezier(DEFPOINTS3D).approximated_length(64)
    assert length == pytest.approx(127.12269127455725)


def test_pickle_support(bezier):
    curve = bezier(DEFPOINTS3D)
    pickled_curve = pickle.loads(pickle.dumps(curve))
    for index in range(3):
        assert pickled_curve.control_points[index] == DEFPOINTS3D[index]


POINTS2D = [
    (0.0, 0.0),
    (1.0, 0.9),
    (2.0, 1.6),
    (3.0, 2.1),
    (4.0, 2.4),
    (5.0, 2.5),
    (6.0, 2.4),
    (7.0, 2.1),
    (8.0, 1.6),
    (9.00, 0.9),
    (10.0, 0.0),
]
TANGENTS2D = [
    (10.0, 10.0),
    (10.0, 8.0),
    (10.0, 6.0),
    (10.0, 4.0),
    (10.0, 2.0),
    (10.0, 0.0),
    (10.0, -2.0),
    (10.0, -4.0),
    (10.0, -6.0),
    (10.0, -8.0),
    (10.0, -10.0),
]

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2011-2023, Manfred Moitzi
# License: MIT License
import ezdxf
from ezdxf.tools.test import load_entities
from ezdxf.sections.objects import ObjectsSection
from ezdxf.entities import Point


def test_load_section():
    doc = ezdxf.new("R2000")
    ent = load_entities(TESTOBJECTS, "OBJECTS")

    section = ObjectsSection(doc, ent)
    assert len(section) == 6
    assert section[0].dxftype() == "DICTIONARY"


class TestAuditObjectSection:
    def test_auditor_removes_invalid_entities(self):
        doc = ezdxf.new()
        count = len(doc.objects)
        # hack hack hack!
        doc.objects._entity_space.add(Point())
        auditor = doc.audit()
        assert len(auditor.fixes) == 1
        assert len(doc.objects) == count, "should call purge() automatically"

    def test_audit_restores_deleted_owner_tag(self):
        doc = ezdxf.new()
        d = doc.rootdict.add_new_dict("TestMe")
        d.dxf.discard("owner")
        doc.audit()
        assert d.dxf.owner == doc.rootdict.dxf.handle, "expected rootdict as owner"

    def test_validate_known_dictionaries(self):
        doc = ezdxf.new()
        materials = doc.rootdict.get_required_dict("ACAD_MATERIAL")
        v1 = materials.add_dict_var("X", "VAR1")
        v2 = materials.add_dict_var("Y", "VAR2")
        assert len(doc.materials) == 5

        auditor = doc.audit()
        assert len(auditor.fixes) == 2
        assert len(doc.materials) == 3
        assert v1.is_alive is False
        assert v2.is_alive is False

    def test_remove_orphaned_dictionary_and_owned_entries(self):
        doc = ezdxf.new()
        orphaned_dict = doc.objects.add_dictionary("ABBA")
        d1 = orphaned_dict.add_dict_var("D1", "VAR1")

        doc.audit()
        assert orphaned_dict.is_alive is False
        assert d1.is_alive is False

    def test_remove_orphaned_dictionary_but_preserve_shared_entries(self):
        doc = ezdxf.new()
        valid_dict = doc.rootdict.add_new_dict("MyDict")
        d1 = valid_dict.add_dict_var("D1", "VAR1")
        orphaned_dict = doc.objects.add_dictionary("ABBA")
        orphaned_dict["D1"] = d1

        doc.audit()
        assert orphaned_dict.is_alive is False
        assert d1.is_alive is True


TESTOBJECTS = """  0
SECTION
  2
OBJECTS
  0
DICTIONARY
  5
C
330
0
100
AcDbDictionary
281
     1
  3
ACAD_COLOR
350
73
  3
ACAD_GROUP
350
D
  3
ACAD_LAYOUT
350
1A
  3
ACAD_MATERIAL
350
72
  3
ACAD_MLEADERSTYLE
350
D7
  3
ACAD_MLINESTYLE
350
17
  3
ACAD_PLOTSETTINGS
350
19
  3
ACAD_PLOTSTYLENAME
350
E
  3
ACAD_SCALELIST
350
B6
  3
ACAD_TABLESTYLE
350
86
  3
ACAD_VISUALSTYLE
350
99
  3
ACDB_RECOMPOSE_DATA
350
499
  3
AcDbVariableDictionary
350
66
  0
DICTIONARY
  5
2A2
330
2
100
AcDbDictionary
280
     1
281
     1
  3
ACAD_LAYERSTATES
360
2A3
  0
DICTIONARY
  5
E6
330
10
100
AcDbDictionary
280
     1
281
     1
  0
DICTIONARY
  5
15D
330
1F
100
AcDbDictionary
280
     1
281
     1
  0
DICTIONARY
  5
28C
330
28B
100
AcDbDictionary
280
     1
281
     1
  3
ASDK_XREC_ANNOTATION_SCALE_INFO
360
28D
  0
DICTIONARY
  5
291
330
290
100
AcDbDictionary
280
     1
281
     1
  3
ASDK_XREC_ANNOTATION_SCALE_INFO
360
292
  0
ENDSEC
"""



# cython: language_level=3
# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License
from typing import Iterable, TYPE_CHECKING, Sequence, Optional
from libc.math cimport fabs, M_PI, M_PI_2, M_PI_4, M_E, sin, tan, pow, atan, log
from .vector cimport (
    isclose, 
    Vec2, 
    v2_isclose, 
    Vec3, 
    v3_sub, 
    v3_add, 
    v3_mul, 
    v3_normalize, 
    v3_cross, 
    v3_magnitude_sqr,
    v3_isclose,
)

import cython

if TYPE_CHECKING:
    from ezdxf.math import UVec

cdef extern from "constants.h":
    const double ABS_TOL
    const double REL_TOL
    const double M_TAU

cdef double RAD_ABS_TOL = 1e-15
cdef double DEG_ABS_TOL = 1e-13
cdef double TOLERANCE = 1e-10

def has_clockwise_orientation(vertices: Iterable[UVec]) -> bool:
    """ Returns True if 2D `vertices` have clockwise orientation. Ignores
    z-axis of all vertices.

    Args:
        vertices: iterable of :class:`Vec2` compatible objects

    Raises:
        ValueError: less than 3 vertices

    """
    cdef list _vertices = [Vec2(v) for v in vertices]
    if len(_vertices) < 3:
        raise ValueError('At least 3 vertices required.')

    cdef Vec2 p1 = <Vec2> _vertices[0]
    cdef Vec2 p2 = <Vec2> _vertices[-1]
    cdef double s = 0.0
    cdef Py_ssize_t index

    # Using the same tolerance as the Python implementation:
    if not v2_isclose(p1, p2, REL_TOL, ABS_TOL):
        _vertices.append(p1)

    for index in range(1, len(_vertices)):
        p2 = <Vec2> _vertices[index]
        s += (p2.x - p1.x) * (p2.y + p1.y)
        p1 = p2
    return s > 0.0

def intersection_line_line_2d(
        line1: Sequence[Vec2],
        line2: Sequence[Vec2],
        bint virtual=True,
        double abs_tol=TOLERANCE) -> Optional[Vec2]:
    """
    Compute the intersection of two lines in the xy-plane.

    Args:
        line1: start- and end point of first line to test
            e.g. ((x1, y1), (x2, y2)).
        line2: start- and end point of second line to test
            e.g. ((x3, y3), (x4, y4)).
        virtual: ``True`` returns any intersection point, ``False`` returns
            only real intersection points.
        abs_tol: tolerance for intersection test.

    Returns:
        ``None`` if there is no intersection point (parallel lines) or
        intersection point as :class:`Vec2`

    """
    # Algorithm based on: http://paulbourke.net/geometry/pointlineplane/
    # chapter: Intersection point of two line segments in 2 dimensions
    cdef Vec2 s1, s2, c1, c2, res
    cdef double s1x, s1y, s2x, s2y, c1x, c1y, c2x, c2y, den, us, uc
    cdef double lwr = 0.0, upr = 1.0

    s1 = line1[0]
    s2 = line1[1]
    c1 = line2[0]
    c2 = line2[1]

    s1x = s1.x
    s1y = s1.y
    s2x = s2.x
    s2y = s2.y
    c1x = c1.x
    c1y = c1.y
    c2x = c2.x
    c2y = c2.y

    den = (c2y - c1y) * (s2x - s1x) - (c2x - c1x) * (s2y - s1y)

    if fabs(den) <= abs_tol:
        return None


    # den near zero is checked by if-statement above:
    with cython.cdivision(True):
        us = ((c2x - c1x) * (s1y - c1y) - (c2y - c1y) * (s1x - c1x)) / den

    res = Vec2(s1x + us * (s2x - s1x), s1y + us * (s2y - s1y))
    if virtual:
        return res

    # 0 = intersection point is the start point of the line
    # 1 = intersection point is the end point of the line
    # otherwise: linear interpolation
    if lwr <= us <= upr:  # intersection point is on the subject line
        with cython.cdivision(True):
            uc = ((s2x - s1x) * (s1y - c1y) - (s2y - s1y) * (s1x - c1x)) / den
        if lwr <= uc <= upr:  # intersection point is on the clipping line
            return res
    return None

cdef double _determinant(Vec3 v1, Vec3 v2, Vec3 v3):
    return v1.x * v2.y * v3.z + v1.y * v2.z * v3.x + \
           v1.z * v2.x * v3.y - v1.z * v2.y * v3.x - \
           v1.x * v2.z * v3.y - v1.y * v2.x * v3.z

def intersection_ray_ray_3d(
        ray1: tuple[Vec3, Vec3],
        ray2: tuple[Vec3, Vec3],
        double abs_tol=TOLERANCE
    ) -> Sequence[Vec3]:
    """
    Calculate intersection of two 3D rays, returns a 0-tuple for parallel rays,
    a 1-tuple for intersecting rays and a 2-tuple for not intersecting and not
    parallel rays with points of closest approach on each ray.

    Args:
        ray1: first ray as tuple of two points as Vec3() objects
        ray2: second ray as tuple of two points as Vec3() objects
        abs_tol: absolute tolerance for comparisons

    """
    # source: http://www.realtimerendering.com/intersections.html#I304
    cdef:
        Vec3 o2_o1
        double det1, det2
        Vec3 o1 = Vec3(ray1[0])
        Vec3 p1 = Vec3(ray1[1])
        Vec3 o2 = Vec3(ray2[0])
        Vec3 p2 = Vec3(ray2[1])
        Vec3 d1 = v3_normalize(v3_sub(p1, o1), 1.0)
        Vec3 d2 = v3_normalize(v3_sub(p2, o2), 1.0)
        Vec3 d1xd2 = v3_cross(d1, d2)
        double denominator = v3_magnitude_sqr(d1xd2)

    if denominator <= abs_tol:
        # ray1 is parallel to ray2
        return tuple()
    else:
        o2_o1 = v3_sub(o2, o1)
        det1 = _determinant(o2_o1, d2, d1xd2)
        det2 = _determinant(o2_o1, d1, d1xd2)
        with cython.cdivision(True):  # denominator check is already done
            p1 = v3_add(o1, v3_mul(d1, (det1 / denominator)))
            p2 = v3_add(o2, v3_mul(d2, (det2 / denominator)))

        if v3_isclose(p1, p2, abs_tol, abs_tol):
            # ray1 and ray2 have an intersection point
            return p1,
        else:
            # ray1 and ray2 do not have an intersection point,
            # p1 and p2 are the points of closest approach on each ray
            return p1, p2

def arc_angle_span_deg(double start, double end) -> float:
    if isclose(start, end, REL_TOL, DEG_ABS_TOL):
        return 0.0

    start %= 360.0
    if isclose(start, end % 360.0, REL_TOL, DEG_ABS_TOL):
        return 360.0

    if not isclose(end, 360.0, REL_TOL, DEG_ABS_TOL):
        end %= 360.0

    if end < start:
        end += 360.0
    return end - start

def arc_angle_span_rad(double start, double end) -> float:
    if isclose(start, end, REL_TOL, RAD_ABS_TOL):
        return 0.0

    start %= M_TAU
    if isclose(start, end % M_TAU, REL_TOL, RAD_ABS_TOL):
        return M_TAU

    if not isclose(end, M_TAU, REL_TOL, RAD_ABS_TOL):
        end %= M_TAU

    if end < start:
        end += M_TAU
    return end - start


def is_point_in_polygon_2d(
    point: Vec2, polygon: list[Vec2], double abs_tol=TOLERANCE
) -> int:
    """
    Test if `point` is inside `polygon`.  Returns +1 for inside, 0 for on the 
    boundary and  -1 for outside.

    Supports convex and concave polygons with clockwise or counter-clockwise oriented
    polygon vertices.  Does not raise an exception for degenerated polygons.


    Args:
        point: 2D point to test as :class:`Vec2`
        polygon: list of 2D points as :class:`Vec2`
        abs_tol: tolerance for distance check

    Returns:
        +1 for inside, 0 for on the boundary, -1 for outside

    """
    # Source: http://www.faqs.org/faqs/graphics/algorithms-faq/
    # Subject 2.03: How do I find if a point lies within a polygon?
    # Numpy version was just 10x faster, this version is 23x faster than the Python 
    # version!
    cdef double  a, b, c, d, x, y, x1, y1, x2, y2
    cdef list vertices = polygon
    cdef Vec2 p1, p2
    cdef int size, last, i
    cdef bint inside = 0

    size = len(vertices)
    if size < 3:  # empty polygon
        return -1
    last = size - 1
    p1 = <Vec2> vertices[0]
    p2 = <Vec2> vertices[last]

    if v2_isclose(p1, p2, REL_TOL, ABS_TOL):  # open polygon
        size -= 1
        last -= 1
    if size < 3:
        return -1

    x = point.x
    y = point.y
    p1 = <Vec2> vertices[last]
    x1 = p1.x
    y1 = p1.y

    for i in range(size):
        p2 = <Vec2> vertices[i]
        x2 = p2.x
        y2 = p2.y

        # is point on polygon boundary line:
        # is point in x-range of line
        a, b = (x2, x1) if x2 < x1 else (x1, x2)
        if a <= x <= b:
            # is point in y-range of line
            c, d = (y2, y1) if y2 < y1 else (y1, y2)
            if (c <= y <= d) and fabs(
                (y2 - y1) * x - (x2 - x1) * y + (x2 * y1 - y2 * x1)
            ) <= abs_tol:
                return 0  # on boundary line
        if ((y1 <= y < y2) or (y2 <= y < y1)) and (
            x < (x2 - x1) * (y - y1) / (y2 - y1) + x1
        ):
            inside = not inside
        x1 = x2
        y1 = y2
    if inside:
        return 1  # inside polygon
    else:
        return -1  # outside polygon

cdef double WGS84_SEMI_MAJOR_AXIS = 6378137
cdef double WGS84_SEMI_MINOR_AXIS = 6356752.3142
cdef double WGS84_ELLIPSOID_ECCENTRIC = 0.08181919092890624
cdef double RADIANS = M_PI / 180.0
cdef double DEGREES = 180.0 / M_PI


def gps_to_world_mercator(double longitude, double latitude) -> tuple[float, float]:
    """Transform GPS (long/lat) to World Mercator.
    
    Transform WGS84 `EPSG:4326 <https://epsg.io/4326>`_ location given as
    latitude and longitude in decimal degrees as used by GPS into World Mercator
    cartesian 2D coordinates in meters `EPSG:3395 <https://epsg.io/3395>`_.

    Args:
        longitude: represents the longitude value (East-West) in decimal degrees 
        latitude: represents the latitude value (North-South) in decimal degrees.

    """
    # From: https://epsg.io/4326
    # EPSG:4326 WGS84 - World Geodetic System 1984, used in GPS
    # To: https://epsg.io/3395
    # EPSG:3395 - World Mercator
    # Source: https://gis.stackexchange.com/questions/259121/transformation-functions-for-epsg3395-projection-vs-epsg3857
    longitude = longitude * RADIANS  # east
    latitude = latitude * RADIANS  # north
    cdef double e_sin_lat = sin(latitude) * WGS84_ELLIPSOID_ECCENTRIC
    cdef double c = pow(
        (1.0 - e_sin_lat) / (1.0 + e_sin_lat), WGS84_ELLIPSOID_ECCENTRIC / 2.0
    )  # 7-7 p.44
    y = WGS84_SEMI_MAJOR_AXIS * log(tan(M_PI_4 + latitude / 2.0) * c)  # 7-7 p.44
    x = WGS84_SEMI_MAJOR_AXIS * longitude
    return x, y


def world_mercator_to_gps(double x, double y, double tol = 1e-6) -> tuple[float, float]:
    """Transform World Mercator to GPS.

    Transform WGS84 World Mercator `EPSG:3395 <https://epsg.io/3395>`_
    location given as cartesian 2D coordinates x, y in meters into WGS84 decimal
    degrees as longitude and latitude `EPSG:4326 <https://epsg.io/4326>`_ as
    used by GPS.

    Args:
        x: coordinate WGS84 World Mercator
        y: coordinate WGS84 World Mercator
        tol: accuracy for latitude calculation

    """
    # From: https://epsg.io/3395
    # EPSG:3395 - World Mercator
    # To: https://epsg.io/4326
    # EPSG:4326 WGS84 - World Geodetic System 1984, used in GPS
    # Source: Map Projections - A Working Manual
    # https://pubs.usgs.gov/pp/1395/report.pdf
    cdef double eccentric_2 = WGS84_ELLIPSOID_ECCENTRIC / 2.0
    cdef double t = pow(M_E, (-y / WGS84_SEMI_MAJOR_AXIS))  # 7-10 p.44
    cdef double e_sin_lat, latitude, latitude_prev

    latitude_prev = M_PI_2 - 2.0 * atan(t)  # 7-11 p.45
    while True:
        e_sin_lat = sin(latitude_prev) * WGS84_ELLIPSOID_ECCENTRIC
        latitude = M_PI_2 - 2.0 * atan(
            t * pow(((1.0 - e_sin_lat) / (1.0 + e_sin_lat)), eccentric_2)
        )  # 7-9 p.44
        if fabs(latitude - latitude_prev) < tol:
            break
        latitude_prev = latitude

    longitude = x / WGS84_SEMI_MAJOR_AXIS  # 7-12 p.45
    return longitude * DEGREES, latitude * DEGREES